#include "AST.h"
#include "CodeContext.h"
#include "CfSource.h"
#include "CtType.h"
#include "MethodDecl.h"
#include "code.h"
#include "code-assign.h"
#include "decls.h"


void FieldDeclNode::codeGen( CodeContext &context )
{
  TypeNode *itype;
  string initializer; 
  if (initExpr()->absent()) {
    if (dtype()->isTitaniumArrayType() || dtype()->isImmutable()) {
      extern const string callNoArgConstructor(TypeNode *t);

      assert(dtype()->isTitaniumArrayType() || dtype()->isImmutable());
      initializer = callNoArgConstructor(dtype());
    } else if (dtype()->isReference())
      initExpr(new CastNode(new NullPntrNode(), dtype()));
    else if (dtype()->isPrimitive())
      initExpr(new CastNode(new LitNode(intLiteral(string("0"))), dtype()));
    else
      assert(0);
    itype = dtype();
  } else
    itype = initExpr()->type();

  if (initializer.empty())
    initializer = initExpr()->emitExpression( context );

  if ((decl()->modifiers() & Static))
    context << makeAssign("&STATIC_REF(" + cFieldName(decl()) + ")",
			  initializer, true, assignKind(itype), true);
  else
    context << MANGLE_THIS_VAR(<<) << "->" << cFieldName(decl()) << " = " << initializer;

  context << ';' << endCline;
}

extern bool opt_dataflow;

void VarDeclNode::codeGen( CodeContext &context )
{
  // We cannot just initialize directly due to C++ restrictions on
  // skipping over initializations when jumping to "case" labels.
  // If you don't believe me, see CD2,  6.7 [stmt.dcl],  3.

  const string varName = MANGLE_STACK_VAR(+, *simpName()->simpName()->ident());
  context.declare( varName, dtype()->cType() );

  string initial;
  bool have_initial = true;

  if (initExpr()->absent()) {
    // from code-alloc.cc
    extern const string callNoArgConstructor(TypeNode *t);

    if (dtype()->isTitaniumArrayType() || dtype()->isImmutable()) {
      initial = callNoArgConstructor(dtype());
    }
    else {
      have_initial = false;
    }
  } 
  else {
    initial = initExpr()->emitExpression( context );
  }

  if (have_initial) {
    context << varName << " = " << initial << ";" << endCline;
  }
  else {
    assert(opt_dataflow);
    context << varName << ";" << endCline;
  }
}


void MethodDeclNode::codeGen( CfSource &os )
{
  MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );

  if (!(mdecl.modifiers() & (Common::Native | Common::Abstract))) {
    bool first = true;
    extern bool line_directives;

    if ((debugger_level > 0) && (line_directives)) {
      const SourcePosn pos = position();
      if (!pos.unknown()) {
	os << "#line " << pos.posnToLineNumber()
	   << " \"" << *(pos.file->name()) << "\"\n";
      }
    }

    os << returnType()->cType() << ' '
       << mdecl.cMethodNameStatic() << '(';
    
    const TypeNode *thisType = mdecl.thisType();
    if (thisType) {
      os << thisType->cType() << " " << MANGLE_THIS_VAR(<<);
      first = false;
    }	

    foriter (param, params()->allChildren(), TreeNode::ChildIter) {
      if (first) first = false; else os << ",";
      os << (*param)->dtype()->cType()
	 << " " << MANGLE_STACK_VAR(<<, *(*param)->simpName()->ident());
    }

    os << ")";
    {
      CodeContext context( os );
      body()->codeGen( context );
    }
    os << '\n';
  }
}


void MethodSignatureNode::codeGen ( CodeContext & )
{
}


void ConstructorDeclNode::codeGen( CfSource &os )
{
  const MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );
  const TypeNode &thisType = *mdecl.thisType();
  const string thisTypeName = thisType.cType().name;
  extern bool line_directives;

  if ((debugger_level > 0) && (line_directives)) {
    const SourcePosn pos = position();
    if (!pos.unknown()) {
      os << "#line " << pos.posnToLineNumber()
	 << " \"" << *(pos.file->name()) << "\"\n";
    }
  }
  
  os << thisTypeName << ' '
     << decl()->cMethodNameStatic()
     << "(";

  bool first = true;
  if (!thisType.isImmutable()) {
    os << thisTypeName << " " << MANGLE_THIS_VAR(<<);
    first = false;
  }

  foriter (param, params()->allChildren(), TreeNode::ChildIter) {
    if (!first) os << ",";
    os << (*param)->dtype()->cType()
       << " " << MANGLE_STACK_VAR(<<, *(*param)->simpName()->ident());
    first = false;
  }
  os << ")";

  {
    CodeContext context( os );
    const MethodDecl &mdecl = static_cast< MethodDecl & >( *decl() );
    const TypeNode &thisType = *mdecl.thisType();
    
    if (thisType.isImmutable()) {
      // declare me
      context.declare( MANGLE_THIS_VAR(+), thisType.cType() );
      // zero out the contents of me
      context << "memset( &" << MANGLE_THIS_VAR(<<) << ", 0, sizeof(" << MANGLE_THIS_VAR(<<) << ") );" << endCline
	      << decl()->container()->cInitFieldsName()
	      << "( &" << MANGLE_THIS_VAR(<<) << " );" << endCline;
    } else {
      constructorCall()->codeGen( context );
      context << endCline;
    }
    
    body()->codeGen( context );
    context << "return " << MANGLE_THIS_VAR(<<) << ";" << endCline;
  }
  os << "\n\n";
}
