/*
  code-defs.cc

  Performs Def/Use analysis.  The first part of the file contains utilities,
  then we have the code for uses, then the code for defs.

  */

#include <cassert>
#include <fstream>
#include <iomanip>
#include <iostream>
#include "utils.h"
#include "AST.h"
#include "decls.h"
#include "errors.h"
#include "interface.h"
#include "code.h"
#include "config.h"
#include "code-grid.h"
#include "code-util.h"
#include "code-defs.h"

extern bool debug_defs;

// Return a canonical representation of the type t with no modifiers.
static xtype type2xtype(TypeNode *t)
{
  TypeNode *c = t->newModifiers((Common::Modifiers) 0);
  return (xtype) c->typeName();
}

static const string& decl2str(Decl *d)
{
  return *d->name();
}

#define type2str(t) ((t)->cType().name)
#define xtype2str(t) ((string) (t))
#define field2str(f) (f)

static void dump_list(ostream &os, llist<DefNode *> *l, int indent)
{
  foreach (f, llist<DefNode *>, *l) {
    (*f)->t->print(os, indent);
    os << '\n';
  }
}

static void dump_field_to_list(ostream &os, const string& prefix,
			       map_string_to_list m, int indent)
{
  for (map_string_to_list::const_iterator f = m.begin();
       f != m.end();
       f++) {
    for (int j = 0; j < indent; j++) os << ' ';
    os << prefix << "(" << field2str((*f).first) << ")=\n";
    dump_list(os, (*f).second, indent + 10);
  }
}

static string plural(int n, const string& s)
{
  return (n == 1) ? "1 " + s : (int2string(n) + ' ' + s + 's');
}

void Defs::print(ostream &os, int indent)
{
  int j;

  for (j = 0; j < indent; j++) os << ' ';
  os << "varDefs:";
  j = 0;
  for (map_decl_to_list::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    os << ' ' << *((*v).first)->name();
    j++;
  }
  if (j == 0) os << " none";
  os << '\n';

  for (j = 0; j < indent; j++) os << ' ';
  os << "type_anyField(type) holds for:";
  j = 0;
  for (map_type_to_bool::const_iterator a = _type_anyField.begin();
       a != _type_anyField.end();
       a++) {
    if ((*a).second) os << ' ' << xtype2str((*a).first);
    j++;
  }
  if (j == 0) os << " none";
  os << '\n';

  for (map_type_field_to_list::const_iterator b = _type_field_to_list.begin();
       b != _type_field_to_list.end();
       b++) {
    dump_field_to_list(os, "type_field_to_list(" + xtype2str((*b).first) + ")",
		       (*b).second, indent);
  }

  for (map_type_to_list::const_iterator d = _arrayDefs.begin();
       d != _arrayDefs.end();
       d++) {
    for (j = 0; j < indent; j++) os << ' ';
    int count = (*d).second->size();
    os << "array type " << xtype2str((*d).first)
      << " has " << plural(count, "arrayDef");
    os << '\n';
  }

  os << "unknownMethods: " << (unknownMethods ? "yes" : "no") << '\n';
}

void Defs::printVardefs(ostream &os, int indent) {
  for (int j = 0; j < indent; j++) os << ' ';
  for (map_decl_to_list::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    os << (*v).first->name()->c_str() << ": ";
    dump_list(cout, (*v).second, 0);
    os << endl;
  }
}

void Defs::printBitset(Bitset *bs, ostream &os, int indent) {
  for (map_decl_to_list::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    ListIterator<DefNode *>iter = (*v).second;
    for (; !iter.isDone(); iter.next()) {
      DefNode *def = *iter;
      if (bs->test(def->index)) {
	def->t->print(os, indent);
	os << endl;
      }
    }
  }
}

Defs::~Defs() {
  for (map_decl_to_list::const_iterator v = _varDefs.begin();
       v != _varDefs.end();
       v++) {
    llist<DefNode *> *deflist = (*v).second;
    while (deflist) {
      delete deflist->front();
      deflist = deflist->free();
    }
  }  
}

/////////////////////////////////////////////////////////////////////////////
// Uses
/////////////////////////////////////////////////////////////////////////////

inline bool Defs::varDefsWithDecl(Decl *decl)
{
  if (debug_defs) {
    cout << "varDefsWithDecl(" << decl2str(decl) << ") -> "
	 << (varDefs(decl) != NULL) << "\n";
  }
  return (varDefs(decl) != NULL);
}

inline bool Defs::fieldDefsWithinType(TypeNode *type)
{
  if (debug_defs) {
    cout << "fieldDefsWithType(" << type2str(type) << ") -> "
	 << type_anyField(type) << "\n";
  }
  return type_anyField(type);
}

void Defs::arrayDef(TreeNode *assignment, TypeNode *t)
{
 llist<DefNode *> *&l = arrayDefs(t);
  l = cons(new DefNode(assignment, nextindex++), l);
  if (debug_defs) {
    cout << "arrayDef at " << assignment->position().asString() << ":\n";
    assignment->print(cout, 4);
    cout << '\n';
  }
}

inline bool Defs::arrayDefsWithinType(TypeNode *type)
{
  return (arrayDefs(type) != NULL);
}

bool Defs::varModified(Decl *decl)
{
  return varDefsWithDecl(decl);
}

#define localOrParam(decl) \
	((decl)->category() & (Decl::LocalVar | Decl::Formal))

bool Defs::varMaybeModified(Decl *decl)
{
  bool vM = varModified(decl);
  bool ret = vM || (!localOrParam(decl) && unknownMethods);

  if (debug_defs) {
    cout << "The thing specified by this decl (" << decl2str(decl) << ") is "
	 << (ret ? "maybe" : "not")
	 << " modified";

    cout << "\n";
    if (true) 
      cout << "  varModifed(decl)=" << vM
	   << "\n  localOrParam(decl)=" << localOrParam(decl)
	   << "\n  unknownMethods=" << unknownMethods
	   << "\n";
  }
  return ret;
}

bool Defs::arrayMaybeModified(TypeNode *type) {
  return arrayDefsWithinType(type);
}

bool Defs::fieldMaybeModified(TypeNode *type, const string &field)
{
  return (type_field_to_list(type, field) != NULL || unknownMethods);
}

bool Defs::fieldNotModified(TypeNode *type, const string &field)
{
  return !fieldMaybeModified(type, field);
}

/////////////////////////////////////////////////////////////////////////////
// Defs
/////////////////////////////////////////////////////////////////////////////

bool NameNode::hasLval() { return true; }

// Returns the instance of the defined variable in this expression, if any.
TreeNode *getDefVar(TreeNode *t) {
  if (isAssignNode(t)) {
    if (isNameNode(t->child(0)->child(0))) {
      return t->child(0)->child(0);
    }
  }
  else if (isVarDeclNode(t)) {
    return t->simpName();
  }
  else if (isForEachPairNode(t)) {
    return t->simpName();
  }
  else if (isParameterNode(t)) {
    return t->simpName();
  }
  return NULL;
}
  

// If t defines a variable, then the def and kill sets given are updated
// accordingly. 
void Defs::merge(TreeNode *t, Bitset *defs, Bitset *kills) {
  TreeNode *defvar = getDefVar(t);
  if (!defvar) {
    return;
  }
  Decl *decl = defvar->decl();
  if (!decl) {
    return;
  }
  llist<DefNode *> *deflist = varDefs(decl);
  //assert(deflist);
  if (!deflist) {
    return;
  }
  ListIterator<DefNode *> iter = deflist;
  for (; !iter.isDone(); iter.next()) {
    DefNode *defnode = *iter;
    if (defnode->t == t) {
      defs->set(defnode->index);
      kills->clear(defnode->index);
    }
    else {
      kills->set(defnode->index);
    }
  }
}

// If t defines a variable, then existing defs of that variable in the
// reaching set are killed and the appropriate def is set.
void Defs::reaching(TreeNode *t, Bitset *reaching) {
  TreeNode *defvar = getDefVar(t);
  if (!defvar) {
    return;
  }
  Decl *decl = defvar->decl();
  if (!decl) {
    return;
  }
  llist<DefNode *> *deflist = varDefs(decl);
  //assert(deflist);
  if (!deflist) {
    return;
  }
  ListIterator<DefNode *> iter = deflist;
  for (; !iter.isDone(); iter.next()) {
    DefNode *defnode = *iter;
    if (defnode->t == t) {
      reaching->set(defnode->index);
    }
    else {
      reaching->clear(defnode->index);
    }
  }
}

// Finds the instance of the defining variable in a given defining expression,
// and hooks up a d->u edge to the given use.
void connectDuEdge(TreeNode *defExpr, TreeNode *use) {
  TreeNode *defVar = getDefVar(defExpr);
  assert(defVar);
  defVar->setUses(cons(use, defVar->getUses()));
}

// Figure out which defs reach the given NameNode (a use) and connect
// u->d edges. 
void Defs::connect(TreeNode *nameNode, Bitset *reaching) {
  Decl *decl = nameNode->decl();
  llist<DefNode *> *deflist = varDefs(decl);
  //assert(deflist);
  if (!deflist) {
    return;
  }
  llist<TreeNode *> *udlist = NULL;
  ListIterator<DefNode *> iter = deflist;
  for (; !iter.isDone(); iter.next()) {
    DefNode *defnode = *iter;
    if (reaching->test(defnode->index)) {
      udlist = cons(defnode->t, udlist);
      connectDuEdge(defnode->t, nameNode);
    }
  }
  nameNode->setDefs(udlist);
}

#if 0
void Defs::classVarAssignment(TreeNode *assignment)
{
  TreeNode *lhs = assignment->child(0);
  llist<TreeNode *> *&l =
    classVarDefs(lhs->ftype(), *lhs->simpName()->ident());

  l = cons(assignment, l);

  if (debug_defs) {
    cout << "classVarDef at " << lhs->position().asString() << ":\n";
    assignment->print(cout, 4);
    cout << '\n';
  }
}
#endif

void Defs::addTypeFieldDef(TypeNode *t, const string& field,
			   TreeNode *assignment)
{
  llist<DefNode *> *&l = type_field_to_list(t, field);
  l = cons(new DefNode(assignment, nextindex++), l);
  type_anyField(t) = true;
  if (debug_defs) {
    cout << "fieldDef < " << t->typeName() << ", " << field << " > at "
	 << assignment->position().asString() << ":\n";
    assignment->print(cout, 4);
    cout << '\n';
  }
}  

/* Note (in this) that the variable specified by d is modified at t. */
void Defs::addVarDef(TreeNode *t, Decl *d)
{
  llist<DefNode *> *&l = varDefs(d);
  l = cons(new DefNode(t, nextindex++), l);
  if (debug_defs) {
    cout << "varDef at " << t->position().asString() << ":\n";
    t->print(cout, 4);
    cout << '\n';
  }
}

void Defs::varAssignment(TreeNode *assignment)
{
  TreeNode *lhs = assignment->child(0);
  if (isNameNode(lhs->child(0))
      && lhs->child(0)->decl() != NULL) {
    addVarDef(assignment, lhs->child(0)->decl());
  } else assert(0);
}

void Defs::assignment(TreeNode *t)
{
  TreeNode *lhs = t->child(0);
  
  if (lhs->isArrayAccessNode()) {
    arrayDef(t, lhs->array()->type());
  }
  else if (isTypeFAN(lhs))
    addTypeFieldDef(lhs->ftype(), *lhs->simpName()->ident(), t);
  else if (isOFAN(lhs))
    addTypeFieldDef(lhs->object()->type(), *lhs->simpName()->ident(), t);
  else if (isObjectNode(lhs))
    varAssignment(t);
  else {
    cerr << "Unknown type of assignment (!):\n";
    t->print(cerr, 4);
    cerr << '\n';
    Error("Defs analysis can't handle unknown type of assignment.");
    assert(0);
  }
}

Defs *TreeNode::defs()
{
  undefined("defs");
  return 0;
}

Defs *ForEachStmtNode::defs()
{
  return _defs;
}

void TreeNode::findDefs(Defs *d)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->findDefs(d);
}

void VarDeclNode::findDefs(Defs *d)
{
  d->addVarDef(this, decl());
}

void ForEachPairNode::findDefs(Defs *d)
{
  d->addVarDef(this, simpName()->decl());
}

void ParameterNode::findDefs(Defs *d)
{
  d->addVarDef(this, simpName()->decl());
}

#define AssignmentDefs(Node)			\
void Node::findDefs(Defs *d)			\
{						\
  d->assignment(this);				\
  foriter (p, allChildren(), ChildIter)		\
    (*p)->findDefs(d);				\
}

AssignmentDefs(AssignNode)
AssignmentDefs(BinaryArithAssignNode)
AssignmentDefs(ShiftAssignNode)
AssignmentDefs(BitwiseAssignNode)
AssignmentDefs(MethodCallAssignNode)
AssignmentDefs(IncrDecrNode)


/* Do findDefs() in all chilren.  Set d->unknownMethods if
   the method is not known to be side effect free. */
void MethodCallNode::findDefs(Defs *d)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->findDefs(d);

  if (d->unknownMethods || isSideEffectFree()) return;

  if (debug_defs) {
    cout << "`unknown' method at " << position().asString() << ":\n";
    print(cout, 4);
    cout << "\n";
  }
  d->unknownMethods = true;
}

