/* code-foreach.cc */

/*

  To get good performance on foreach over RectDomains, we need to
  strength reduce foreach ourselves because of the division in the
  index expression. RectDomains can have non-unit stride, so the
  division must be there in a naive implementation.  C compilers then
  refuse to do strength reduction on the address calculation even when
  the addresses are in an arithmetic progression.

  Idea:

  Before code gen, identify index expressions that are candidates for
  strenth reduction.  E.g., A[p] where p is the iteration variable of
  a foreach.  Mark the index expression and the foreach.

  At code gen of a foreach, check every item on the list of possibly
  strength-reducible index expressions.  If the expression is simply
  A[p] then we need A to be loop invariant; if the expression is
  A[j*p+k] then we need A and j and k to be loop invariant; etc.
  There are further complications in checking validity, though.  The
  problem is that the difference between &A[p0] and &A[p1] may not be
  integral for adjacent iterations p0 and p1.  This only comes up when
  A[p] is not used on every iteration.  See the bottom of the file for
  a copy of a relevent email from May, 1997.

  If it is indeed reducible, we create a pointer and a
  pointer-delta-per-iteration for it in the code output.

  If all uses of the iteration variable are optimized, then we may omit
  the code to create the point object.

  All of the above could be applied to the rects comprising a general
  Domain, but it is less clear whether it is worthwhile.

*/

#include <cassert>
#include <fstream>
#include <iomanip>
#include <map>
#include <set>
#include <stack>
#include "AST.h"
#include "CodeContext.h"
#include "CtGlobal.h"
#include "CtLocal.h"
#include "CtReference.h"
#include "CtType.h"
#include "PrimitiveDecl.h"
#include "code-assign.h"
#include "code-commonsub.h"
#include "code-defs.h"
#include "code-foreach.h"
#include "code-grid.h"
#include "code-poly.h"
#include "code-util.h"
#include "code.h"
#include "config.h"
#include "ctBox.h"
#include "decls.h"
#include "delimit.h"
#include "domain-decls.h"
#include "errors.h"
#include "interface.h"
#include "lgMacro.h"
#include "osstream.h"
#include "utils.h"

class Candidate;

typedef map< TreeNode *, TreeNode *, less < TreeNode * > > map_tree_to_tree;
typedef map< TreeNode *, map_tree_to_tree, less < TreeNode * > >
    map_tree_tree_to_tree; 
typedef map< void *, ExprNode *, less < void * > > map_void_to_expr;
typedef map< void *, Candidate *, less < void * > > map_void_to_cand;
typedef map< TreeNode *, map_void_to_expr, less < TreeNode * > >
    map_tree_void_to_expr; 
typedef map< void *, int, less< void * > > map_void_to_int;
typedef map< void *, map_void_to_int, less< void * > > map_void_void_to_int;
typedef map< int, llist<Candidate *> *, less< int > > map_expr_to_class;
typedef set< const TreeNode *, less< const TreeNode * > > treeSet;
typedef map< TreeNode *, treeSet *, less < TreeNode * > > map_tree_to_treeSet;

#ifdef HAVE_BINARY_ADAPTORS
typedef stack< string, list< string > > StringStack;
#else
typedef stack<         list< string > > StringStack;
#endif


int SRTempNum = 0;

#define DEBUG_OSR debug_opt
#define DEBUG_FOREACH_OPT debug_opt

#ifndef DEBUG_INDUCTION_VAR
#define DEBUG_INDUCTION_VAR debug_opt
#endif

#ifndef DEBUG_LOOP_INVAR
#define DEBUG_LOOP_INVAR debug_opt
#endif

/////////////////////////////////////////////////////////////////////////////

static map_tree_tree_to_tree sums;

static inline void foundSumOrDifference(TreeNode *t, TreeNode *WRTloop,
				 TreeNode *invariantPart)
{
  sums[t][WRTloop] = invariantPart;
}

static inline bool isSum(TreeNode *t, TreeNode *WRTloop)
{
  return (arithCall(t) == '+' && sums[t][WRTloop] != NULL);
}

static inline bool isDifference(TreeNode *t, TreeNode *WRTloop)
{
  return (arithCall(t) == '-' && sums[t][WRTloop] != NULL);
}

static inline TreeNode *baseOfSum(TreeNode *t, TreeNode *WRTloop)
{
  TreeNode *invariantPart = sums[t][WRTloop];
  TreeNode *left = t->child(0)->child(0);
  TreeNode *right = t->child(1)->child(0);
  
  return (left == invariantPart) ? right : left;
}

static inline TreeNode *offsetOfSum(TreeNode *t, TreeNode *WRTloop)
{
  return sums[t][WRTloop];
}

/////////////////////////////////////////////////////////////////////////////

class Candidate {
public:  
  TreeNode *accessNode, *WRTloop;

  /* If not NULL, this candidate is exactly equal to another one, and we should
     not duplicate variables and computation for this. */
  Candidate *identical;

  /* ID for temporaries in the generated code.  All non-identical index
     exprs have different tempNums.  Zero means the candidate is invalid. */
  int tempNum;

  /* List of Candidates expressed as offsets off this. */
  llist<Candidate *> *candsOffsetOffThis;

  /* String naming the variable that holds the array used by this
     candidate. */
  string arrayStr;

  /* If non-null, this candidate will be expressed as a constant plus
     the given candidate. */
  Candidate *offsetOffExpr;

  Candidate(TreeNode *acc, TreeNode *foreachstmtnode) {
    accessNode = acc;
    WRTloop = foreachstmtnode;
    tempNum = 0;
    candsOffsetOffThis = NULL;
    offsetOffExpr = NULL;
  }

  bool shared() { return (identical != NULL); }

  TreeNode *index() { return accessNode->index(); }
  TreeNode *array() { return accessNode->array(); }
  bool isSum() { return ::isSum(index(), WRTloop); }
  bool isDifference() { return ::isDifference(index(), WRTloop); }
  bool isSumOrDifference() { return isSum() || isDifference(); }
  TreeNode *base() { return ::baseOfSum(index(), WRTloop); }
  TreeNode *offset() { return ::offsetOfSum(index(), WRTloop); }

  bool isLocal() const {
    return accessNode->array()->type()->isLocal();
  }

  bool isLocalInferred() const {
    return accessNode->array()->type()->isLocalInferred();
  }

  bool valid(void) {
    return (tempNum > 0);
  }
  string value(CodeContext &context) {
    /* SR_x has multiple incarnations, one per C loop nest.  So a loop of
       arity N should use SR_x_(N-1). oSR has only one incarnation, though. */ 

    if (shared()) return identical->value(context);
    if (offsetOffExpr) {
      const CtType &elemType = accessNode->type()->cType();
      const CtReference &addrType = ctBox( elemType,
					   isLocal() );
      const string result = ExprNode::declareTemporary(context, addrType);

      context << lgMacro("INDEX", isLocal()) << '('
	      << result << ", "
	      << "SR_" << offsetOffExpr->tempNum << '_'
	      << WRTloop->vars()->child(0)->initExpr()->type()->tiArity() - 1
	      << ", oSR_" << tempNum << ");" << endCline;
      return result;
      
    } else
      return (string("SR_") + int2string(tempNum)
	      + '_' + int2string(WRTloop->vars()->child(0)->initExpr()
			      ->type()->tiArity() - 1));
  }
  void incrCode(ostream &, int dim);

//   void initCode(ostream &os, string q, int dim, int arity,
// 		const string iterationVar);
};


static void updateIterationPoint(CodeContext &os,
				 string q, string iterationVar, int arity)
{
  os << MANGLE_TI_DOMAINS_POINT_CONSTRUCT(<<, int2string(arity)) << MANGLE_STACK_VAR(<<, iterationVar);
  for (int point_arity = 0; point_arity < arity; point_arity++)
    os << ", " << q << '_' << point_arity;
  os << ");" << endCline;
}

static void bumpPoint(CodeContext &os, string q, int dim, int arity,
		      string iterationVar)
{
  os << endCline << "/* bump " << dim << " */" << endCline;

  os << q << "_" << dim << " += " << q << "d_" << dim << ";" << endCline;
  updateIterationPoint(os, q, iterationVar, arity);
  os << q << "_" << dim << " -= " << q << "d_" << dim << ";" << endCline;
}

void Candidate::incrCode(ostream &out, int dim)
{
  out << lgMacro("INDEX", isLocal()) << '('
      <<  "SR_" << tempNum << '_' << dim << ", "
      <<  "SR_" << tempNum << '_' << dim << ", "
      << "dSR_" << tempNum << '_' << dim << ");"
      << endCline;
}

/////////////////////////////////////////////////////////////////////////////


void TreeNode::foreachOpt0()
{
  foriter (p, allChildren(), ChildIter) {
    (*p)->foreachOpt0();
  }
}

/* Returns whether b is an ancestor of a in the tree. */
static bool treeMember(const TreeNode *a, const TreeNode *b)
{
  while (a != NULL)
    if (a == b) return true;
  else
    a = a->parent();

  return false;
}

/* Takes a NameNode for a point that is a foreach iteration var.
   Returns the ForEachStmtNode that introduced that point. */
static inline const TreeNode *point2foreachNode(const TreeNode *p)
{
  TreeNode *q = p->decl()->source()->parent()->parent();
  assert(isForEachStmtNode(q));
  return q;
}

/* Takes an ArrayAccessNode.  Returns whether it is a grid access. */
#define isGridAccess(t) ((t)->array()->type()->isTitaniumArrayType())
			 
/* Given two NameNodes for points that are foreach iteration vars,
   one foreach much have the other in its body.  Return the point
   corresponding to the inner of the two foreach's. */
static const TreeNode *inner(const TreeNode *a, const TreeNode *b)
{
  if (a == NULL || a == b) return b;
  if (b == NULL) return a;
  if (treeMember(point2foreachNode(a), point2foreachNode(b))) return a;
  return b;
}
static const TreeNode *innermostPoint(TreeNode *t)
{
  if (t == NULL)
    return NULL;
  else if (isNameNode(t)
	   && t->decl()->hasSource()
	   && t->decl()->source() != NULL
	   && isForEachPairNode(t->decl()->source()))
    return t;
  else {
    const TreeNode *ret = NULL;
    foriter (p, t->allChildren(), TreeNode::ChildIter)
      ret = inner(innermostPoint(*p), ret);
    return ret;
  }
}

/* acc is the ArrayAccessNode; fp is the ForEachPairNode that
   introduced the innermost iteration variable used in the index part of
   acc.  This routine adds to the list of potentially optimizable
   index exprs at fp and marks acc with a pointer to the added list
   entry. */
static void mark(TreeNode *acc, TreeNode *fp)
{
  assert(acc->isArrayAccessNode());
  assert(isForEachPairNode(fp));
  if (false /* DEBUG_FOREACH_OPT */) {
    cout << "mark()\n  acc=\n";
    acc->print(cout, 6);
    cout << "\n  fp=\n";
    fp->print(cout, 6);
    cout << '\n';
  }
  TreeNode *f = fp->parent()->parent();
  assert(isForEachStmtNode(f));
  acc->innerForeachIterationVar() = f->candidateIndexExprs() =
    cons(new Candidate(acc, f), f->candidateIndexExprs());
}


llist<Candidate *> *& TreeNode::candidateIndexExprs()
{
  static llist<Candidate *> *bogus = NULL;
  undefined("candidateIndexExprs");
  return bogus;
}

llist<Candidate *> *& ForEachStmtNode::candidateIndexExprs()
{
  return _candidateIndexExprs;
}

llist<Candidate *> *& TreeNode::innerForeachIterationVar()
{
  static llist<Candidate *> *bogus = NULL;
  undefined("innerForeachIterationVar");
  return bogus;
}

llist<Candidate *> * const & TreeNode::innerForeachIterationVar() const
{
  static llist<Candidate *> *bogus = NULL;
  undefined("innerForeachIterationVar");
  return bogus;
}

llist<Candidate *> *& ArrayAccessNode::innerForeachIterationVar()
{
  return _innerForeachIterationVar;
}

llist<Candidate *> * const & ArrayAccessNode::innerForeachIterationVar() const
{
  return _innerForeachIterationVar;
}

bool& TreeNode::needPoint()
{
  static bool bogus = true;
  undefined("needPoint");
  return bogus;
}

bool& ForEachPairNode::needPoint()
{
  return _needPoint;
}

/* fp is a ForEachPairNode whose point is needed by the loop body. */
static void foundNeededPoint(TreeNode *fp)
{
  if (DEBUG_FOREACH_OPT) {
    if (!fp->needPoint()) {
      cout << "Need point:\n";
      fp->print(cout, 2);
      cout << '\n';
    }
  }
  fp->needPoint() = true;
}

/* If we reach here, it means this is a NameNode that is not inside a
   candidate.  Therefore, if the NameNode is a foreach iteration variable,
   assume we need to construct it when we codegen the foreach. */
void NameNode::foreachOpt0()
{
  TreeNode *fp;
  if (decl()->hasSource()
      && (fp = decl()->source()) != NULL
      && isForEachPairNode(fp)
      && parent() != fp) {
    foundNeededPoint(fp);
  }
}

/* If we reach here, it means this is a NameNode that is inside a
   candidate.  The argument to this function is the candidate's
   ForEachPairNode.  Therefore, if the NameNode is a different foreach 
  iteration variable, assume we need to construct it when we codegen
   that foreach. */
void NameNode::needPointAnal(TreeNode *candidate_fp)
{
  TreeNode *fp;
  if (decl()->hasSource()
      && (fp = decl()->source()) != NULL
      && candidate_fp != fp
      && isForEachPairNode(fp)
      && parent() != fp) {
    foundNeededPoint(fp);
  }
}

void TreeNode::needPointAnal(TreeNode *candidate_fp)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->needPointAnal(candidate_fp);
}

void ArrayAccessNode::foreachOpt0()
{
  if (isGridAccess(this)) {
    const TreeNode *z = innermostPoint(index());
    if (z != NULL) {
      mark(this, z->decl()->source());
      needPointAnal(z->decl()->source());
      return;
    }
  }
  array()->foreachOpt0();
  index()->foreachOpt0();
}

void foreachOptPass()
{
  foreach (f, llist<TreeNode *>, *allFiles) 
    (*f)->foreachOpt0();
}

/////////////////////////////////////////////////////////////////////////////

static bool allInvariant(TreeNode *expr, TreeNode *loop)
{
  foriter (p, expr->allChildren(), TreeNode::ChildIter)
    if (!((*p)->loopInvariant(loop)))
      return false;

  return true;
}

/////////////////////////////////////////////////////////////////////////////

llist<TreeNode *> *& TreeNode::invariants()
{
  static llist<TreeNode *> *bogus = NULL;
  undefined("invariants");
  return bogus;
}

llist<TreeNode *> *& ForEachStmtNode::invariants()
{
  return _invariants;
}

/* Return whether t appears on the list of invariants for WRTloop. */
static bool isInvar(const TreeNode *t, TreeNode *WRTloop)
{
  static map_tree_to_treeSet m;
  treeSet *& s = m[WRTloop];
  if (!s) {
    s = new treeSet;
    foreach (f, llist<TreeNode *>, *WRTloop->invariants())
      s->insert((const TreeNode *) *f);
  }
  return (s->count(t) > 0); 
}

static void foundInvariant(TreeNode *inv, TreeNode *WRTloop)
{
  assert(isForEachStmtNode(WRTloop));
  ForEachStmtNode *l = (ForEachStmtNode *) WRTloop;
  if (DEBUG_LOOP_INVAR) {
    cout << "found invariant (WRTloop " << WRTloop->position().asString()
      << "):\n";
    inv->print(cout, 6);
    cout << '\n';
  }
  l->invariants() = cons(inv, l->invariants());
}

/////////////////////////////////////////////////////////////////////////////

bool TreeNode::loopInvariant(TreeNode *WRTloop)
{
  foriter (p, allChildren(), ChildIter) {
    (*p)->loopInvariant(WRTloop);
  }
  return false;
}

static bool methodTakesLoopInvariantsToLoopInvariant(TreeNode *t)
{
  return t->isPureFunction();
}

bool MethodCallNode::loopInvariant(TreeNode *WRTloop)
{
  if (arithCall(this)) {
    TreeNode *left = child(0)->child(0);
    TreeNode *right = child(1)->child(0);
    bool lefti = left->loopInvariant(WRTloop);
    bool righti = right->loopInvariant(WRTloop);
    if (lefti && righti && tiOrNumericType(left) && tiOrNumericType(right)) {
      foundInvariant(this, WRTloop);
      return true;
    }
    if (DEBUG_LOOP_INVAR) {
      cout << "non-invariant arithCall (lefti=" << lefti
	   << " righti=" << righti << "):\n";
      print(cout, 12);
      cout << '\n';
    }
  } else if (methodTakesLoopInvariantsToLoopInvariant(this)
	     && (isTypeFAN(child(0))
		 || isOFAN(child(0))
		 && child(0)->child(0)->loopInvariant(WRTloop))) {
    bool b = true;
    foriter (p, args()->allChildren(), ChildIter) {
      if (!(*p)->loopInvariant(WRTloop)) b = false;
    }
    if (b) {
      foundInvariant(this, WRTloop);
      return true;
    }
  }
  
  foriter (p, allChildren(), ChildIter) {
    (*p)->loopInvariant(WRTloop);
  }
  return false;
}

bool LitNode::loopInvariant(TreeNode *WRTloop)
{
  foundInvariant(this, WRTloop);
  return true;
}

bool CastNode::loopInvariant(TreeNode *WRTloop)
{
  if (opnd0()->loopInvariant(WRTloop)) {
    foundInvariant(this, WRTloop);
    return true;
  }
  return false;
}


bool ArrayAccessNode::loopInvariant(TreeNode *WRTloop)
{
  if (DEBUG_LOOP_INVAR) {
    cout << "checking if array access is loop invariant: ";
    print(cout, 0);
    cout << ")\n";
  }
  if (allInvariant(this, WRTloop)) {
    if (WRTloop->defs()->arrayMaybeModified(array()->type())) { 
      return false;
    }
    foundInvariant(this, WRTloop);
    return true;
  } 
  return false;
}

bool PointNode::loopInvariant(TreeNode *WRTloop)
{
  if (allInvariant(child(0), WRTloop)) {
    foundInvariant(this, WRTloop);
    return true;
  }
  return false;
}

bool FieldAccessNode::loopInvariantField(TypeNode *type,
					 const string *field,
					 TreeNode *WRTloop)
{
  if (WRTloop->defs()->fieldNotModified(type, *field)) {
    foundInvariant(this, WRTloop);
    return true;
  }
  return false;
}

bool ObjectFieldAccessNode::loopInvariant(TreeNode *WRTloop)
{
  return object()->loopInvariant(WRTloop)
    && loopInvariantField(object()->type(), simpName()->ident(), WRTloop);
}

bool TypeFieldAccessNode::loopInvariant(TreeNode *WRTloop)
{
  return loopInvariantField(ftype(), simpName()->ident(), WRTloop);
}

bool ThisNode::loopInvariant(TreeNode *WRTloop)
{
  foundInvariant(this, WRTloop);
  return true;
}

bool ObjectNode::loopInvariant(TreeNode *WRTloop)
{
  if (name()->loopInvariant(WRTloop)) {
    foundInvariant(this, WRTloop);
    return true;
  }
  return false;
}

bool NameNode::loopInvariant(TreeNode *WRTloop)
{
  if (DEBUG_LOOP_INVAR) {
    cout << "checking if namenode is loop invariant: ";
    print(cout, 0);
    cout << ")\n";
  }
  if (isForEachPairNode(parent())) return false;
  if (qualifier()->absent()) {
    if (isObjectNode(parent())
	&& WRTloop->defs()->varNotModified(parent()->decl())) {
      assert(!treeMember(parent()->decl()->source(), WRTloop));
      foundInvariant(this, WRTloop);
      return true;
    }
  }

  return false;
}

/////////////////////////////////////////////////////////////////////////////

void TreeNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  foriter (p, allChildren(), ChildIter)
    (*p)->liftInvariants(context, WRTloop);
}

void ExprNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  static map_tree_void_to_expr m;

  foriter (p, allChildren(), ChildIter)
    (*p)->liftInvariants(context, WRTloop);

  if (!isInvar(this, WRTloop)) return;

  if (DEBUG_LOOP_INVAR) {
    cout << "Lifting a loop invariant expression:\n";
    print(cout, 4);
    cout << '\n';
  }

  ExprNode *& lifted = m[WRTloop][commonSubExpr()];
  if (lifted) cVariable = lifted->cVariable;
  else {
    lifted = this;
    // If not already lifted from an enclosing loop, code gen it now
    if (cVariable == "") {
#if 0
      // tends to cause tcbuild cache misses
      context << "#if 0" << endCline;
      print(context, 4);
      context << endCline << "#endif" << endCline;
#endif
      cVariable = simpleVar(context);
    }

    if (DEBUG_LOOP_INVAR)
      cout << "Lifted to C variable " << cVariable << ".\n";
  }
}

/* The same as any other ExprNode, except that we take care not to
   lift a value that is a method.  A method call is fine, of
   course. */
void FieldAccessNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  if (parent() == NULL || !isMethodCallNode(parent()))
    ExprNode::liftInvariants(context, WRTloop);
}


/* The same as any other ExprNode, except that strength reduced index
   expressions are optimized in a different way and it would be
   pointless to lift them.  Consider lifting their parts, though. */
void ArrayAccessNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  if (isStrengthReduced()) {
    array()->liftInvariants(context, WRTloop);
    // index()->liftInvariants(context, WRTloop);
  } else
    ExprNode::liftInvariants(context, WRTloop);
}

/* In the common case, use the normal ExprNode::liftInvariants().
   In the case of certain trivial methods, don't bother to lift it, but
   recurse on the guts in case something interesting is there. */
void MethodCallNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  if (isGridMethod_domain(this)) {
    method()->liftInvariants(context, WRTloop);
    args()->liftInvariants(context, WRTloop);
  } else
    ExprNode::liftInvariants(context, WRTloop);
}

void ObjectNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  // Probably too trivial to lift---could hurt performance (but could help it).
}

void LitNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  // Too trivial to lift---in fact, doing so may increase register pressure
  // and hurt the performance.
}

void ThisNode::liftInvariants(CodeContext &context, TreeNode *WRTloop)
{
  // Too trivial to lift---in fact, doing so may increase register pressure
  // and hurt the performance.
}

/////////////////////////////////////////////////////////////////////////////
// analysis related to induction variables
/////////////////////////////////////////////////////////////////////////////

llist<TreeNode *> *& TreeNode::inductionVars()
{
  static llist<TreeNode *> *bogus = NULL;
  undefined("inductionVars");
  return bogus;
}

llist<TreeNode *> *& ForEachStmtNode::inductionVars()
{
  return _inductionVars;
}

static bool isInductionVar(TreeNode *t, TreeNode *WRTloop)
{
  foreach (f, llist<TreeNode *>, *WRTloop->inductionVars())
    if (*f == t) return true;

  return false;
}

static void foundInductionVar(TreeNode *ind, TreeNode *WRTloop)
{
  assert(isForEachStmtNode(WRTloop));
  ForEachStmtNode *l = (ForEachStmtNode *) WRTloop;
  if (DEBUG_INDUCTION_VAR) {
    cout << "found induction variable:\n";
    ind->print(cout, 6);
    cout << '\n';
  }
  l->inductionVars() = cons(ind, l->inductionVars());
}

bool TreeNode::inductionVar(TreeNode *WRTloop)
{
  foriter (p, allChildren(), ChildIter) {
    (*p)->inductionVar(WRTloop);
  }
  return false;
}

/* If this is loop invariant, don't bother.  Otherwise, if all pieces
   are loop invariants or induction variables, then this is an
   inductionVar. */
bool PointNode::inductionVar(TreeNode *WRTloop)
{
  if (isInvar(this, WRTloop)) return false;

  bool ok = true;
  foriter (p, args()->allChildren(), ChildIter)
    if (!isInvar(*p, WRTloop)
	&& (!(*p)->inductionVar(WRTloop)))
      ok = false;

  if (ok)
    foundInductionVar(this, WRTloop);

  return ok;
}

bool ArrayAccessNode::inductionVar(TreeNode *WRTloop)
{
  bool arr = array()->inductionVar(WRTloop);
  bool ind = index()->inductionVar(WRTloop);
  if (arr && array()->type()->isPointType()
      && (ind || isInvar(index(), WRTloop))) {
    foundInductionVar(this, WRTloop);
    return true;
  }
  
  return false;
}

bool MethodCallNode::inductionVar(TreeNode *WRTloop)
{
  char op = arithCall(this);

  if (op) {
    TreeNode *left = child(0)->child(0);
    TreeNode *right = child(1)->child(0);
    bool lefti = left->inductionVar(WRTloop);
    bool righti = right->inductionVar(WRTloop);
    if (lefti && righti)
      if (op == '-' || op == '+') {
	foundInductionVar(this, WRTloop);
	return true;
      } else return false;
    if ((lefti || righti)
	&& (op == '-' || op == '+' || op == '*')
	&& ((lefti || isInvar(left, WRTloop))
	    && (righti || isInvar(right, WRTloop)))) {
      foundInductionVar(this, WRTloop);
      if (op == '+' || op == '-')
	foundSumOrDifference(this, WRTloop, (righti ? left : right));
      return true;
    }

    return false;
  }

  foriter (p, allChildren(), ChildIter) {
    (*p)->inductionVar(WRTloop);
  }
  return false;
}

bool ObjectNode::inductionVar(TreeNode *WRTloop)
{
  if (name()->inductionVar(WRTloop)) {
    foundInductionVar(this, WRTloop);
    return true;
  } else return false;
}

bool NameNode::inductionVar(TreeNode *WRTloop)
{
  if (decl()->hasSource()
      && decl()->source() != NULL
      && decl()->source()->parent()->parent() == WRTloop) {
    foundInductionVar(this, WRTloop);
    return true;
  }

  return false;
}

/////////////////////////////////////////////////////////////////////////////

static bool arrayIsInvariant(TreeNode *arrayAccess, TreeNode *WRTloop)
{
  return isInvar(arrayAccess->array(), WRTloop)
    || (cheating && 
	(isObjectNode(arrayAccess->child(0))
	 && isNameNode(arrayAccess->child(0)->child(0))
	 || isOFAN(arrayAccess->child(0))
	 || isTypeFAN(arrayAccess->child(0))
	 || isThisFAN(arrayAccess->child(0))));
}

static bool indexExprIsInductionVar(TreeNode *arrayAccess, TreeNode *WRTloop)
{
  return isInductionVar(arrayAccess->index(), WRTloop);
}

/* Destructively remove invalid candidates from l. */
static llist<Candidate *> *exciseInvalidCandidates(llist<Candidate *> *l)
{
  if (l == NULL) return NULL;
  else {
    l->tail() = exciseInvalidCandidates(l->tail());
    return (l->front()->valid()) ? l : l->tail();
  }
}

/* Go through the candidate index expressions and decide which ones
we're going to strength reduce.  Return true iff we're going to
strength reduce all of them.  Destructively remove candidates we aren't going
to optimize. */
static bool filterCandidates(ostream &os, TreeNode *t)
{
  map_void_to_cand previousExpressions;
  int opt = 0, unopt = 0, redundant = 0;
  bool aii, ieiiv;
  llist<Candidate *> *l = t->candidateIndexExprs();
  foreach (cand, llist<Candidate *>, *l) {
    TreeNode *arrayAccess = (*cand)->accessNode;
    stat_foreach_index_exprs++;
    if ((aii = arrayIsInvariant(arrayAccess, t))
	& (ieiiv = indexExprIsInductionVar(arrayAccess, t))) {
      if (DEBUG_FOREACH_OPT) 
	cout << "Checking whether we have seen this index expr before...\n";
      Candidate *& prev =
	previousExpressions[arrayAccess->commonSubExpr()];
      if (prev == NULL) {
	prev = *cand;
	(*cand)->tempNum = ++SRTempNum;
	(*cand)->identical = NULL;
      } else {
	(*cand)->tempNum = prev->tempNum;
	(*cand)->identical = prev;
	stat_redundant_index_exprs++;
	redundant++;
      }

      opt++;
      stat_foreach_index_exprs_optimized++;

      if (DEBUG_FOREACH_OPT) {
	cout << "Found "
	     << (!(*cand)->shared() ? "an index expr to strength reduce ("
		 : "a redundant index expr (")
	     << opt << ")\n";
	arrayAccess->print(cout, 4);
	cout << '\n';
      }
    } else {
      unopt++;
      if (DEBUG_FOREACH_OPT) {
	cout << "Couldn't optimize index expr because ";
	if (!aii)
	  cout << "array isn't loop inv.";
	if (!aii && !ieiiv)
	  cout << " and ";
	if (!ieiiv)
	  cout << "index expr is not an ind. var";
	cout << '\n';
	arrayAccess->print(cout, 4);
	cout << '\n';
      }
    }
  }
  if (opt > 0) stat_foreach_optimized++;
  os << "/* foreach at " << t->position().asString() << " has "
     << opt << " of " << (opt + unopt) << " index exprs optimized ("
     << redundant << " redundant) */" << endCline;

  t->candidateIndexExprs() = exciseInvalidCandidates(t->candidateIndexExprs());

  return (unopt == 0);
}

/****************************************************************************
 Offset Strength Reduction

 View all expressions as one of three types: p + i, p, or p - i, where
 p is an induction var and i is loop invariant.  Classify all
 expressions with the same p together.  The difference between any two
 expressions classified together is then a loop invariant.  So, for
 example, we might generate code that looks like a for p and (a + b)
 for p + i.  Then we need only increment a at the bottom of the loop,
 rather than two variables.
 ***************************************************************************/

// Of the expressions with the same p, find the "simplest."
// If p alone is available, use that.  Otherwise, just pick something.
static Candidate *simplestOffset(llist<Candidate *> *l)
{
  foreach (cand, llist<Candidate *>, *l)
    if (!(*cand)->isSumOrDifference()) return *cand;

  return l->front();
}

static void annotateSimplestOffset(Candidate *simplest, llist<Candidate *> *cl)
{
  if (DEBUG_FOREACH_OPT) {
    cout << "OSR: simplest is:\n";
    simplest->accessNode->print(cout, 4);
    cout << "\nOSR: less simple are:\n";
  }

  foreach (cand, llist<Candidate *>, *cl) 
    if (simplest != *cand) {
      simplest->candsOffsetOffThis = cons(*cand, simplest->candsOffsetOffThis);
      (*cand)->offsetOffExpr = simplest;
      ++stat_foreach_osr;

      if (DEBUG_FOREACH_OPT) {
	(*cand)->accessNode->print(cout, 4);
	cout << '\n';
      }
    }
}

static inline int OSRclassification(Candidate *cand)
{
  static int unique = 0;
  static map_void_void_to_int memo;
  if (DEBUG_OSR)
    cout << "begin OSRClassification().\n";
  int &ret = memo[cand->array()->commonSubExpr()]
    [(cand->isSumOrDifference()
      ? cand->base()
      : cand->index())->commonSubExpr()];
  if (ret == 0) ret = ++unique;
  if (DEBUG_OSR)
    cout << "end OSRClassification(): return value is " << ret << "\n.";
  return ret;
}

static void offsetStrengthReduce(ostream &os, TreeNode *t)
{
  map_expr_to_class m;
  llist<Candidate *> *l = t->candidateIndexExprs();
  int i = 0;

  if (DEBUG_OSR)
    cout << "\nOSR for foreach at " << t->position().asString() << ":\n\n";

  // View all expressions as one of three types: p + i, p, or p - i,
  // where is an induction var and i is loop invariant.
  // Classify all expressions with the same p together.
  foreach (cand, llist<Candidate *>, *l)
    if (!(*cand)->shared()) {
      int q = OSRclassification(*cand);
      m[q] = cons(*cand, m[q]);
    }

  // Of the expressions with the same p, find the "simplest."
  // If p alone is available, use that.  Otherwise, just pick something.
  for (map_expr_to_class::const_iterator u = m.begin(); u != m.end(); u++) {
    int size = (*u).second->size();
    if (DEBUG_FOREACH_OPT) {
      cout << "OSR: class " << ++i << " has size "
	   << size << ".\n";
      if (size == 1) {
	(*u).second->front()->accessNode->print(cout, 4);
	cout << '\n';
      }
    }
    if (size > 1) {
      Candidate *simplest = simplestOffset((*u).second);
      annotateSimplestOffset(simplest, m[OSRclassification(simplest)]);
    }
  }
}

/////////////////////////////////////////////////////////////////////////////

static void
foreachRectDomainStart(CodeContext &context, int arity, string q, TreeNode *dom,
		       string& curr_rect)
{
  // RectDomain specific startup
  if (isObjectNode(dom) || isOFAN(dom) || isTypeFAN(dom)) {
    curr_rect = dom->emitExpression(context);
  } else if (isGridMethod_domain(dom)) {
    curr_rect = dom->child(0)->child(0)->emitExpression(context) + ".domain";
  } else {
    string exp = dom->emitExpression(context);
    curr_rect = q + "_curr_rect";
    context.declare(curr_rect, RectDomainNDecl[arity - 1]->cType());
    context << curr_rect << " = " << exp << ";" << endCline;
  }
}


/* BUG: this code should use context.declare(...) to declare variables. */
static void
foreachGeneralStart(CodeContext &context, int arity, string q, TreeNode *dom)
{
  // General domain specific startup
  // iter = d.getRectangles().newIterator()
  // do curr_rect = iter[i]

  ClassDecl &d = *DomainNDecl[ arity - 1 ];
  ClassDecl &rdl = *RectDomainListNDecl[ arity - 1 ];
  ClassDecl &iter = *RectDomainListIteratorNDecl[ arity - 1 ];

  const CtType &d_desc = d.cDescriptorType();
  const CtType &rdl_desc = rdl.cDescriptorType();
  const CtType &iter_desc = iter.cDescriptorType();  

  const CtGlobal &gl_d_desc = *new CtGlobal( *new CtLocal( d_desc ) );
  const CtGlobal &gl_rdl_desc = *new CtGlobal( *new CtLocal( rdl_desc ) );
  const CtGlobal &gl_iter_desc = *new CtGlobal( *new CtLocal( iter_desc ) );

  const CtGlobal &d_inst = *new CtGlobal( d.cType() );
  const CtGlobal &rdl_inst = *new CtGlobal( rdl.cType() );
  const CtGlobal &iter_inst = *new CtGlobal( iter.cType() );

  const string d_var = q + "_d";
  const string mdl_var = q + "_mdl";
  const string d_iter_var = q + "_d_iter";
  const string iter_cont_var = q + "_iter_cont";
  const string d_desc_var = q + "_d_desc";
  const string mdl_desc_var = q + "_mdl_desc";
  const string d_iter_desc_var = q + "_d_iter_desc";
  const string curr_rect_var = q + "_curr_rect";

  context.depend( d_desc );
  context.depend( rdl_desc );
  context.depend( iter_desc );

  context.depend( gl_d_desc );
  context.depend( gl_rdl_desc );
  context.depend( gl_iter_desc );

  context.declare( d_var, d_inst );
  context.declare( mdl_var, rdl_inst );
  context.declare( d_iter_var, iter_inst );
  context.declare( iter_cont_var, PrimitiveDecl::BoolDecl.cType() );

  context << d_desc    << " *" << d_desc_var << ';'      << endCline
	  << rdl_desc  << " *" << mdl_desc_var << ';'    << endCline
	  << iter_desc << " *" << d_iter_desc_var << ';' << endCline;
    
  string exp = dom->emitExpression(context);

  context << d_var << " = " << exp << ';' << endCline
	  << "CLASS_INFO_GLOBAL(" << d_desc_var << ", "
	  << d_desc << ", " << d_var << ");" << endCline
	
	  << mdl_var << " = " << d_desc_var << "->" << MANGLE_TI_DOMAINS_DOMAIN_GETRECTANGLES << d_var << ");" << endCline
	  << "CLASS_INFO_GLOBAL(" << mdl_desc_var << ", "
	  << rdl_desc << ", " << mdl_var << ");" << endCline
	
	  << d_iter_var << " = " << mdl_desc_var << "->" << MANGLE_TI_DOMAINS_RDL_NEWITERATOR << mdl_var << ");" << endCline
	  << "CLASS_INFO_GLOBAL(" << d_iter_desc_var << ", "
	  << iter_desc << ", " << d_iter_var << ");" << endCline

	  << "do {" << endCline
	  << "  " << iter_cont_var << " = !(" << d_iter_desc_var << "->" << MANGLE_TI_DOMAINS_RDL_ISEND << d_iter_var << "));" << endCline
	  << "  if (" << iter_cont_var << ") {" << endCline
	  << "     RD" << arity << " " << curr_rect_var << ";" << endCline
	  << "     " << curr_rect_var << " = " << d_iter_desc_var << "->" << MANGLE_TI_DOMAINS_RDL_DEREFERENCE << d_iter_var << ");" << endCline;
}

static StringStack pointGuts(CodeContext& context, TreeNode *pt)
{
  StringStack q;
  int arity = pt->type()->tiArity();

  if (isPointNode(pt)) {
    for (int i = 0; i < arity; i++)
      q.push(pt->args()->child(i)->emitExpression(context));
  } else {
    string index0 = pt->simpleVar(context);
    for (int i = 0; i < arity; i++)
      q.push(index0 + ".x" + int2string(i));
  }
  return q;
}

#define quickconv(indices, arr) quickconv2((indices), (arr), true)
#define quickconvDiff(indices, arr) quickconv2((indices), (arr), false)

static string quickconv2(StringStack indices,
			 const string& arr, bool subtractBase)
{
  string s;
  for (int i = indices.size(); i >= 1; i--, indices.pop()) {
    s += "(" + indices.top();
    if (subtractBase) s += "-" + arr + ".base[" + int2string(i - 1) + "]";
    s += ")/" + arr + ".stride[" + int2string(i - 1) + "]*"
      + arr + ".sideFactors[" + int2string(i - 1) + "]"
      + (i == 1 ? "" : "+");
  }
  return '(' + s + ')';
}

#define DIFF(dim, n) (string("dSR_") + int2string(n) + '_' + int2string(dim))
#define OSR_DIFF(n) (string("oSR_") + int2string(n))
#define nonTrivialSize(dim) q << "_" << (dim) << " + " << q << "d_" << (dim) << " < " << q << "e_" << (dim)

static void SRInitArrays(CodeContext &context, llist<Candidate *> *candidates)
{
  foreach (cand, llist<Candidate *>, *candidates) {
    Candidate *c = *cand;
    c->arrayStr = c->array()->simpleVar(context);
  }
}

static void SRInit(CodeContext &context, llist<Candidate *> *candidates,
		   string q, int dim, int arity, const string iterationVar)
{
  bool noneHandled = true;
  map< Candidate *, string, less < Candidate * > > convFirstPointM,
    arrayM;

  /* Part 1: Compute initial value for the SR pointers. */

  foreach (cand, llist<Candidate *>, *candidates) {
    Candidate *c = *cand;
    assert(c->valid());
    if (!c->shared()) {
      int n = c->tempNum;
      TreeNode *accessNode = c->accessNode;
      TypeNode *t = accessNode->array()->type();
      bool isGlobal = !(t->modifiers() & Common::Local);
      const CtType &eltType = t->elementType()->cType();
      const string arityStr = int2string(t->tiArity());
      const string array = c->arrayStr;
      const string R = arrayMethodPrefix(t) + "domain("
	  + eltType + ", " + arityStr + ")(" + array + ")";
      const string prevbase = string("SR_") + int2string(n) + '_' + int2string(dim - 1);
      const string base = string("SR_") + int2string(n) + '_' + int2string(dim);
      string convFirstPoint;
  
      // Only need set this to its initial value once per foreach loop
      if (dim == 0 && noneHandled)
	updateIterationPoint(context, q, iterationVar, arity);

      noneHandled = false;

      if (dim == 0) {
	if (!c->offsetOffExpr)
	  context.declare(base, ctBox(eltType, !isGlobal));
	
	if (c->offsetOffExpr) {
	  // Two cases: In either case, assume the structure of c is of the
	  // form (p + j), for an induction var p and a loop invariant j.
	  if (c->offsetOffExpr->isSumOrDifference())
	    // case 1: offset off of an expr that is of the form (p + i)
	    // compute osr by taking conv(p+j) - conv(p+i)
	    convFirstPoint = quickconv(pointGuts(context, accessNode->index()),
				       array);
	  else
	    // case 2: offset off of an expr that is of the form p
	    // instead of computing osr = conv(p+j) - conv(p), we will
	    // later just compute osr = convdiff(j).
	    convFirstPoint = "";
	} else {
	  StringStack indices = pointGuts(context,
					  accessNode->index());
	  for (int d = 0; d < arity; d++)
	    context.declare(DIFF(d, n), theIntType->cType());

	  convFirstPoint = string("fSR_") + int2string(n);
	  context.declare(convFirstPoint, theIntType->cType());
	  context << convFirstPoint << " = "
		  << quickconv(indices, array) << ";" << endCline;
	  context << lgMacro("INDEX", t->isLocal()) << '('
		  << base << ", " << array << ".A, "
		  << convFirstPoint << ");" << endCline;
	}

	arrayM[c] = array;
	convFirstPointM[c] = convFirstPoint;
      } else {
	// dim > 0
	if (!c->offsetOffExpr)
	  context << ctBox(eltType, !isGlobal) << ' '
		  << base << " = " << prevbase << ";" << endCline;
      }
    }
  }

  if (noneHandled) return;

  /* Part 2 */

  // Compute the (pointer) increments necessary for each dimension.
  // Output that calculation inside an "if (nontrivialsize()) { ... }".
  // If the size of the domain is trivial then we don't need the delta
  // per iteration because there will only be zero or one iterations.
  // Also accumulate the offset strength reduction calculations (into os_tail)
  // and output them at the end, outside the if.  They are always needed.
  if (dim == 0) {
    ostringstream os_tail_;
    {
      CodeContext os_tail(context, os_tail_);
      for (int d = 0; d < arity; d++) {
	if (use_nonTrivialSizeCheck)
	  context << "if (" << nonTrivialSize(d) << ") {";
	bumpPoint(context, q, d, arity, iterationVar);
	foreach (cand, llist<Candidate *>, *candidates) {
	  Candidate *c = *cand;
	  assert(c->valid());
	  if (!c->shared()) {
	    int n = c->tempNum;
	    if (c->offsetOffExpr) {
	      if (d == 0) {
		context.declare(OSR_DIFF(n), theIntType->cType());
		if (convFirstPointM[c] == "") {
		  assert(c->isSumOrDifference());
		  string convOffset =
		    quickconvDiff(pointGuts(os_tail, c->offset()),
				  arrayM[c]);
		  os_tail << OSR_DIFF(n) << " = "
			  << (c->isDifference() ? "-" : "")
			  << convOffset;
		} else {
		  os_tail << OSR_DIFF(n) << " = "
			  << convFirstPointM[c]
			  << " - " << convFirstPointM[c->offsetOffExpr];
		}
		os_tail << ";" << endCline << endCline;
	      }
	    } else {
	      string array = arrayM[c];
	      string convFirstPoint = convFirstPointM[c];
	      string convSecondPoint =
		quickconv(pointGuts(context, c->accessNode->index()), array);
	      context << DIFF(d, n) << " = "
		      << convSecondPoint << " - " << convFirstPoint << ";"
		      << endCline << endCline;
	    }
	  }
	}
	if (use_nonTrivialSizeCheck)
	  context << "}" << endCline; // matches if (nontrivialsize(d)) 
      }
    }
    if (!os_tail_.str().empty())
      context << "/* Offset Strength Reduction (OSR) calculations: */"
	      << endCline
	      << os_tail_.str()
	      << "/* end of OSR calculations */" << endCline;
  }
}
#undef DIFF
#undef OSR_DIFF
#undef nonTrivialSize

static void SRIncr(ostream &out, llist<Candidate *> *candidates, int dim)
{
  foreach (cand, llist<Candidate *>, *candidates)
    if ((*cand)->valid() && !(*cand)->shared() && !(*cand)->offsetOffExpr)
      (*cand)->incrCode(out, dim);
}

#define middle_label(i) "middle" << SRTempNum << "_" << (i)
#define top_label(i) "top" << SRTempNum << "_" << (i)

static void foreachRectIterate(CodeContext &context, int arity, string q,
			       TreeNode *dom, TreeNode *var, TreeNode *stmt,
			       llist<Candidate *> *candidates,
			       bool needPoint, string curr_rect)
{
  // iterate over RectDomain curr_rect
  // s = getLoopingStride;
  // p0 = getLowerBound;
  // p1 = getUpperBound;
  // for (var0, p0.get(0), p1.get(0), s.get(0))
  //   ... for (var-n, p0.get(n), p1.get(n), s.get(n))
  //     do stuff
  //   ... end for var-n
  // end for var0

  int i;

  ++SRTempNum; /* Ensures unique labels for gotos (if any) */

  if (!needPoint) {
    stat_foreach_point_removed++;
    context << "/* foreach iteration point optimized away */" << endCline;
  }

  // preliminaries

  // Pull the rectdomain variables directly from the struct.
  // Warning: This is relies on knowledge about the domain library
  // implementation.
  context << "P" << arity << " " << MANGLE_STACK_VAR(<<, *(var->ident())) << ";" << endCline;
  for (i = 0; i < arity; i++) {
    context << "jint " << q << "_" << i << " = " << curr_rect << "." << MANGLE_TI_DOMAINS_RECTDOMAIN_FIELD_ACCESS(<<, "p0", int2string(arity)) << ".x" << i << ';' << endCline;
    if (i > 0)
      context << "jint " << q << "s_" << i << " = " << q << "_" << i << ";" << endCline;
    context << "jint " << q << "d_" << i << " = " << curr_rect << "." << MANGLE_TI_DOMAINS_RECTDOMAIN_FIELD_ACCESS(<<, "loopStride", int2string(arity)) << ".x" << i << ';' << endCline;
    context << "jint " << q << "e_" << i << " = " << curr_rect << "." << MANGLE_TI_DOMAINS_RECTDOMAIN_FIELD_ACCESS(<<, "p1", int2string(arity)) << ".x" << i << ';' << endCline;
  }

  // If the start and end values are all 0, this is a null domain.
  // Warning: This is relies on knowledge about the domain library
  // implementation.
  context << "if (";
  for (i = 0; i < arity; i++) {
    if (i != 0)
      context << " || ";
    context << "((" << q << "_" << i << " != 0) || ";
    context << "(" << q << "e_" << i << " != 0))";
  }
  context << ") {" << endCline;

  SRInitArrays(context, candidates);
  // top of the loop
  for (i = 0; i < arity; i++) {
    SRInit(context, candidates, q, i, arity, *(var->ident()));

    if (i == 0)
      context << "/* Top of the loop */" << endCline;

    // if i is zero we skip this, because the FOOs_BAR variable doesn't
    // exist---the correct value is already in FOO_BAR.
    if (i > 0)
      context << q << "_" << i << " = " << q << "s_" << i << ";" << endCline;

    if (use_goto) {
      context << "goto " << middle_label(i) << ";" << endCline;
      context << top_label(i) << ":" << endCline;
      SRIncr(context, candidates, i);
      context << middle_label(i) << ":" << endCline;
    } else {
      context << "do {" << endCline;
    }
  }

  // body of the loop
  if (needPoint)
    updateIterationPoint(context, q, *(var->ident()), arity);
  stmt->codeGen(context);

  // bottom of the loop
  for (i = arity; --i >= 0; ) {
    if (use_goto)
      context << "if ((" << q << "_" << i << " += " << q << "d_" << i << ") < "
	 << q << "e_" << i << ") goto " << top_label(i) << ";" << endCline;
    else {
      context << "if ((" << q << "_" << i << " += " << q << "d_" << i << ") >= "
	 << q << "e_" << i << ") break;" << endCline;
      SRIncr(context, candidates, i);
      context << "} while (1);" << endCline;
    }
  }
  context << "}" << endCline;
}

#undef middle_label
#undef top_label


static void foreachGeneralEnd(CodeContext &os, int arity, string q)
{
  // General domain specific tail
  // advance curr_rect
  // end while curr_rect
  os << "     " << q << "_d_iter_desc->" << MANGLE_TI_DOMAINS_RDL_ADVANCE
     << q << "_d_iter);" << endCline;
  os << "  }" << endCline;
  os << "} while (" << q << "_iter_cont);" << endCline;
}

void ForEachStmtNode::emitStatement( CodeContext &context )
{
  context.setDebugLine(this);

  static int tempCount;
  if (SRTempNum == 0) tempCount = 0;
  string q = string("f") + int2string(tempCount++);
  TreeNode *var = vars()->child(0)->simpName();
  TreeNode *dom = vars()->child(0)->initExpr();
  int arity = dom->type()->tiArity();
  bool allValid, isGeneralDomain =
    dom->type()->isDomainType() && !dom->type()->isRectDomainType();

  stat_foreach_count++;

  if (opt_foreach) {
    if (opt_dataflow) {
      //polyAnalyze(this);
    }

    if (debug_opt)
      cout << "\nAnalyzing foreach at " << position().asString() << ":\n\n";

    assert(_defs == NULL);
    findDefs(_defs = new Defs);
    if (debug_defs) {
      cout << "Summary of defs analysis for foreach at "
	   << position().asString() << ":\n";
      _defs->print(cout, 0);
      cout << "(end of summary of defs analysis)\n";
    }

    // The domain over which we are iterating is only evaluated once, so
    // it is safe to consider it a loop invariant.
    foundInvariant(dom, this);

    stmt()->loopInvariant(this);
    stmt()->inductionVar(this);
    allValid = filterCandidates(context, this);

    if (opt_osr)
      offsetStrengthReduce(context, this);

    if (opt_lift) {
      context << "/* lifted loop invariants: */" << endCline;
      liftInvariants(context, this);
      context << "/* end of lifted loop invariants. */" << endCline;
    }
  }
  else context << endCline << "/* foreach */" << endCline;

  {
    CodeContext subcontext( context );

    string curr_rect;
    if (isGeneralDomain) {
      foreachGeneralStart(subcontext, arity, q, dom);
      curr_rect = q + "_curr_rect";
    } else {
      foreachRectDomainStart(subcontext, arity, q, dom, curr_rect);
    }

    {
      CodeContext bodyContext( subcontext );
      foreachRectIterate(bodyContext, arity, q, dom, var, stmt(),
			 candidateIndexExprs(),
			 !opt_foreach || test_sr || !allValid
			 || vars()->child(0)->needPoint(),
			 curr_rect);
    }

    if (isGeneralDomain)
      foreachGeneralEnd(subcontext, arity, q);
  }
  context << "/* hcaerof */" << endCline;
}

/////////////////////////////////////////////////////////////////////////////
// What follows used to be in code-array.cc.
/////////////////////////////////////////////////////////////////////////////

bool ArrayAccessNode::isStrengthReduced() const
{
  return innerForeachIterationVar()
    && innerForeachIterationVar()->front()->valid();
}


const string ArrayAccessNode::strengthReducedValue(CodeContext &context) const
{
  return innerForeachIterationVar()->front()->value(context);
}


/*
  Further discussion of the non-integral increment problem (see
  comment at top of file):

Date: Fri, 30 May 1997 15:17:29 -0700 (PDT)
From: Geoff Pike <pike@priam.CS.Berkeley.EDU>
To: titanium-group@cs.Berkeley.EDU
In-reply-to: <338E6139.6601283B@cs.berkeley.edu> (message from Alex Aiken on
	Thu, 29 May 1997 22:10:17 -0700)
Subject: foreach, foreach, foreach

Further analysis of foreach over RectDomains.

There are two efficiency issues: bounds checks and non-integral increments.
By deciding the bounds checking errors are ERRORS and deciding that
for efficiency we may compile under the assumption that there are no
out-of-bounds array references, much of the mess of two weeks ago is gone.

What are the cases we can handle optimally?  In short, the cases where
every index expression A[e] that appears in the loop body is known to
be valid in every iteration of the loop.  Or A[e] simply appears in
every iteration and is assumed to be valid because bounds checking is
turned off.  Or A[e] appears in every iteration and is proven to be
valid in the loop header as part of bounds checking.  Also, a few
other cases.

Notation:
    p is a point (the control point of the iteration)
    R and Q are RectDomains
    c, d, and k are loop invariant points or integers
    A and B are Titanium arrays (assume 1D for simplicity)

1a. foreach (p in R) { ... A[p] ... } where we know R <= A.domain
(or foreach (p in Q) { ... A[c*p+d] ... } where c*Q+d <= A.domain)

If you want the performance guarantee, one way to make sure you get it
is to only use A[p] for a p running over a domain known to be a subset
of A.domain.  Everything one can do with array foreach falls into this
category, because "array foreach (f in A) ... f ..." is equivalent to
"foreach (p in A.domain) ... A[p] ..."

Why is it fast?  We know A[p] is valid for every p, hence it is safe
and bounds checking is unnecessary.  Furthermore, &A[p] is integral if
A[p] is valid, so the memory increment used in strength reduction is
the difference of two valid addresses, an integer.

This is the big winner.  It is trivial to have programmers understand it.
Also, my feeling is that 99% of loops will fall in this category or 1b.
HOWEVER, THE PROGRAMMER MAY HAVE TO NUDGE THE COMPILER TO MAKE SURE IT
KNOWS WHAT IT NEEDS TO KNOW.

The things listed below are gravy (and increasing obscure).

1b. foreach (p in R) { ... A[c*p+d] ...} // c and d are loop invariant
        where A[c*p+d] occurs in all iterations

(In these examples we need not know a thing about R and A.)

Example 1:  foreach (p in R) { A[p] = 7; }
Example 2*: foreach (p in R) { if (foo()) A[p] = 7; else A[p+3] = 9; }
Example 3:  foreach (p in R) { A[k*p]; if (foo()) A[k*p] = 7; }

Intuition: Assume the program does not have bounds checking errors
(i.e., bounds checking is turned off or we checked it somehow).  If
A[p] is used in every iteration it is valid in every iteration and
non-integral memory increments are unnecessary.  Example 2* may be a
problem.  Example 3 is OK because A[k*p] is appears (and hence assumed to
be valid) in every iteration.  Note that the programmer might write
this way just to be extra sure that the compiler sees &A[k*p] as an
induction variable with integral increment.

If bounds checking is turned off this is really just like case 1a,
i.e., the compiler knows that the domain of access is a RectDomain
and assumes it is a subset of the array's domain because it assumes no
bounds checking errors.

2. foreach (p in R) { ... A[p] ...}
   where (stride of R) times (memory stride of A) times (elt size of A) is
     an integer

This is a subset of the dangerous cases that happen to be no problem
because the granularity of pointers in the hardware is often 4x or 8x
different from the granularity of addresses of interest.  Example 4:
in memory we have
    -----------------------------------------
     ...  | A[0] | A[2] | A[4] | A[6] | ...   (memory stride = 1:2)
    -----------------------------------------

    int i = 0;
    foreach (p in [lo : hi]) { if (bar(i++)) A[p] = 11.0; }

In this example, we can treat A[p] as an induction variable that is
incremented by (memory stride) * (stride of [lo : hi], namely, 1) * (elt size)
each iteration, which comes out to (elt size)/2, which is known at
compile time to be an integer.

Also, sometimes it may be possible to infer that an increment much be
integral even without exact knowledge of the strides.  Compare this
to example 3:

Example 5: foreach (p in R) { A[p] = sqrt(bar()); if (foo()) A[k*p] = 7; }

Here we know that A[p] is always valid, so it's increment is integral.
But A[k*p]'s increment is just k times A[p]'s increment, so it too is integral.

A similar case is where the memory stride of A is somehow known to be
an integer, but we know not what it is.  That's good enough, though.

-----------------------------------------------------------------------------

Now, the bad cases.

Example 6*: foreach (p in R) { A[floor(p*sin(p[2]))] = 11; }

This will compile into something at best equivalent to what C or
FORTRAN would do, and probably worse because we won't be able to
remove the division from the address calculation.

Example 2*: foreach (p in R) { if (foo()) A[p] = 7; else A[p+3] = 9; }

A[p] and A[p+3] can be treated as induction variables, but they may in
general have non-integral increments.  This is the problem pointed out
by David.  Let's simplify the code to just:

Example 7*: foreach (p in R) { if (foo()) A[p] = 7; }

and ask how to compile it.  One solution is not to bother treating
&A[p] as an induction variable.  Then we do the full address
calculation (in general including some multiplies and adds and at
least one divide) each time foo() is true.  The other solutions treat
&A[p] as an induction variable.  Then there are two relevant subcases:
   i.   We don't know whether the increment is integral
   ii.  We know it is not integral
 ( iii. We know it is integral --- see "good cases" above)

For the first case I propose we'll generate something like:
    {
	int index = ..., index_stop = ...;
	if (index < index_stop) {
	    int index_inc = ..., w, inc, when_inc;
	    if (...) { when_inc = ...; w = ...; inc = ...; } // sad
	    else { when_inc = 1; w = 0; inc = ...; }         // happy 
	    T *_a = ...;
	    do {
	        if (foo()) *_a = 7;
		index += index_inc;
	        if (!(index < index_stop)) break;
	        if (++w == when_inc) { w = 0; _a += inc; }
		// or, the last line could be replaced with:
		// if (happy || (++w == when_inc && ((w = 0), true))) _a += inc
	    } while (true);
	}
    }

In case (ii) we'll generate similar but slightly simpler code.

By treating &A[p] as an induction variable we pay a few extra cycles
per iteration instead of paying a bunch of extra cycles per use. 


Geoff

*/
