#include "utils.h"
#include "AST.h"
#include "decls.h"
#include "cfg.h"
#include "code-util.h"
#include "code-defs.h"

void setupUdMethod(MethodDeclNode *t);
static void printUsesBblock(Bblock *b);
static bool debugOn = false;

void NameNode::setDefs(llist<TreeNode *> *defs) {
  _deflist = defs;
}

llist<TreeNode *> *NameNode::getDefs(void) {
  return _deflist;
}

void NameNode::setUses(llist<TreeNode *> *uses) {
  _uselist = uses;
}

llist<TreeNode *> *NameNode::getUses(void) {
  return _uselist;
}

#if 0				// not currently used by anyone
static bool isComplexStmt(TreeNode *t) {
  return (t->isStatementNode() && !isMethodCallNode(t)) || isTreeListNode(t);
}
#endif

void setupUdFile(TreeNode *t) {
  const int childCount = t->arity();

  if (isMethodDeclNode(t)) {
    setupUdMethod((MethodDeclNode *)t);
  }
  else {
    for (int sweep = 0; sweep < childCount; sweep++) {
      // Need to quit the recursion when it's obvious there isn't
      // going to be a MethodDeclNode here.
      setupUdFile(t->child(sweep));
    }
  }  
}

static Defs *methodDefs;

// If this cfg defines a variable, then enter it into the def map.
void mapDefCfg(CfgNode *cfg) { 
  cfg->astNode()->findDefs(methodDefs);
}

// Creates a def map for a given method, which maps definitions to vector
// indices.
void mapDefs(TreeNode *method) { 
  assert(isMethodDeclNode(method));
  if (isOmittedNode(method->body())) {
    return;
  }
  CfgNode *cfg = method->body()->getCfgExtent()->front();
  CfgNode *end = LastMethodNode(method);
  assert(cfg);
  assert(end);
  methodDefs = new Defs();
  UnmarkCfgExts(cfg, end);
  TraverseCfgExtents(cfg, end, mapDefCfg);
}

// Calculates reaching in, reaching out, and the flow function for the block.
// These are accessed as b->defs(), b->defsIn(), and b->defsThis(), 
// respectively.
bool setupDefsBblock(Bblock *b) {
  // Traverse the bblock, setting up defs for all the non-stmt nodes.
  // The stmt nodes should not be traversed because their children are already
  // represented in the CFG. We might have the same problem with children of
  // any node at all, but the lowering of the graph by temporary creation 
  // will prevent defs from appearing as children of non-stmt nodes.

  if (debugOn) {
    cout << "Processing bblock#" << b->number << endl;
  }

  // Calculate the union of the predecessor defs.
  Bitset *ins = new Bitset(methodDefs->size());
  ListIterator<Bblock *> bbIt = b->predIter();
  for (; !bbIt.isDone(); bbIt.next()) {
    Bblock *bb = *bbIt;
    if (bb->defsOut()) {
      ins->un(bb->defsOut());
    }
  }
  
  // If the ins are the same as last time, we don't need to process this
  // block any further.
  if (b->defsIn()) {
    if (b->defsIn()->equal(ins)) {
      delete ins;
      return false;
    }
    else {
      delete b->defsIn();
    }
  }
  b->setDefsIn(ins);

  // If we don't know what the flow function is for this block, figure it out.
  if (!b->defs()) {
    Bitset *defs = new Bitset(methodDefs->size());
    Bitset *kills = new Bitset(methodDefs->size());
    ListIterator<CfgNode *> cfgIt = b->cfgIter();
    for (; !cfgIt.isDone(); cfgIt.next()) {
      CfgNode *cfg = *cfgIt;
      TreeNode *t = cfg->astNode();
      methodDefs->merge(t, defs, kills);
    }
    b->setDefs(defs);
    b->setKills(kills);
  }

  // Determine the defs reaching out: (ins U defs) - kills
  Bitset *outs = new Bitset(methodDefs->size());
  outs->un(ins);
  outs->un(b->defs());
  outs->subtract(b->kills());
  if (b->defsOut()) {
    delete b->defsOut();
  }
  b->setDefsOut(outs);
  return true;
}


// Sets up u->d edges for the uses in this bblock. 
void setupUdBblock(Bblock *b) {
  Bitset *reaching = new Bitset(methodDefs->size());

  reaching->un(b->defsIn());
  // Traverse the CFG list for this block, hooking up u->d edges from each 
  // NameNode that we see. We also need to repeat the dataflow for this 
  // bblock, to allow for def-use pairs within the bblock.
  ListIterator<CfgNode *> cfgIt = b->cfgIter();
  for (; !cfgIt.isDone(); cfgIt.next()) {
    CfgNode *cfg = *cfgIt;
    TreeNode *t = cfg->astNode();
    methodDefs->reaching(t, reaching);
    if (isNameNode(t) && t->decl()) {
      methodDefs->connect(t, reaching);
    }
  }  
  delete reaching;
  if (debugOn) {
    printUsesBblock(b);
  }
}

// An action for the bblock traversal.
void printBlock(Bblock *b) {
  b->print(cout);
}

void setupUdMethod(MethodDeclNode *t) {

  // Methods without stmts don't have CFGs.
  if (!t->getBblockRoot()) {
    return;
  }

  debugOn = true;
  //if (!strcmp(t->decl()->name()->c_str(), "begin")) {
  //    debugOn = true;
  //}
  //  else {
  //    return;
  //  }

  // First, create a global bitvector map for all the defs in this method.
  assert (!methodDefs);
  methodDefs = new Defs;
  t->findDefs(methodDefs);
  //  mapDefs(t);
  
  // Calculate the reaching def sets for all the bblocks in the method.
  TraverseBblocksIter(t->getBblockRoot(), setupDefsBblock);

  // Then go through each block again and hook up the u->d edges.
  TraverseBblocks(t->getBblockRoot(), setupUdBblock);

  // Free the global map.
  delete methodDefs;
  methodDefs = NULL;

  debugOn = false;
}

void printDefList(llist<TreeNode *> *deflist) {
  ListIterator<TreeNode *>listIt = ListIterator<TreeNode *>(deflist);

  for (; !listIt.isDone(); listIt.next()) {
    TreeNode *t = *listIt;
    while (!t->isStatementNode() && !(isVarDeclNode(t))) {
      t = t->parent();
    }
    t->print(cout,1);
    cout << endl;
  }
}

static void printUsesBblock(Bblock *b) {
  printf("-----------------------\n");
  printf("Uses for bblock #%d\n", b->number);
  ListIterator<CfgNode *> cfgIt = b->cfgIter();
  for (; !cfgIt.isDone(); cfgIt.next()) {  
    CfgNode *cfg = *cfgIt;
    TreeNode *t = cfg->astNode();
    if (isNameNode(t)) {
      printf("%s:\n", t->ident()->c_str());
      llist<TreeNode *> *defList= ((NameNode *)t)->getDefs();
      printDefList(defList);
    }
  }
}

#if 0				// not currently used by anyone
static void printDefsBblock(Bblock *b) {
  Bitset *outs = b->defsOut();
  printf("-----------------------\n");
  printf("Defs reaching out of bblock #%d\n", b->number);
  methodDefs->printBitset(outs, cout, 0);
}
#endif
