#include <limits.h>
#include "ti-gc.h"
#include "regions.h"

/* Page allocator for region-based memory management */
/* TBD: coalescing */
/* TBD: special free list for size == K */

struct page
{
  /* low order bit is 1 if previous page is allocated, 0 if not */
  struct page *next;
  /* number of pages in this allocation unit. Field does not exist if
     pagecount == 1 (ie pages list in struct allocator) */
  PageId pagecount;
  /* Only in free pages not in the single_pages list */
  struct page *previous;
};

/* next is the value for the next field */

struct page *alloc_single_page(struct page *next);
void free_single_page(Region r, struct page *p);

struct page *alloc_pages(int n, struct page *next);
void free_pages(Region r, struct page *p);


/* a list of free individual pages */
struct page *single_pages;

/* a list of pages of size < K (only useful for single_pages) */
struct page *solitary_pages;

/* all free pages */
struct page *unused_pages;


/* Doubly linked page list management */
void addfront(struct page **list, struct page *p)
/* Effects: Adds p to the front of doubly-linked list list */
{
  p->previous = NULL;
  p->next = *list;
  if (*list) (*list)->previous = p;
  *list = p;
}

void unlink_page(struct page **list, struct page *p)
/* Effects: Remove p from its doubly linked list */
{
  if (p->previous)
    p->previous->next = p->next;
  else
    *list = p->next;
  if (p->next)
    p->next->previous = p->previous;
}

/* Some memory allocation routines */

#ifdef USE_GC_NONE
#define MINIMUM_MEM_REQUEST (RPAGESIZE * K * K * K)
/* Get memory and align it. */

void *nogc_region_get_mem(size_t s)
{
  size_t request_bytes;
  void *mem;

  /* Don't get less than K * RPAGESIZE extra memory (K * RPAGESIZE
     is the minimum size for something on unused_pages) */
  if (s + K * RPAGESIZE < MINIMUM_MEM_REQUEST) 
    request_bytes = MINIMUM_MEM_REQUEST;
  else
    request_bytes = s;

  mem = malloc(request_bytes + RPAGESIZE - 1);
  mem = (void *)ALIGN((IntPtr)mem, RPAGESIZE);

  /* Add the extra memory to unused_pages */
  if (request_bytes > s)
    {
      struct page *extra = (struct page *)((char *)mem + s);

#ifndef NMEMDEBUG
      set_region_range(extra, (char *)mem + request_bytes, FREEPAGE);
#endif
      extra->pagecount = (request_bytes - s) >> RPAGELOG;
      addfront(&unused_pages, extra);
    }

  return mem;
}
#else /* USE_GC_NONE */
/* The GC uses this routine to get memory */

char *get_titanium_memory(GC_word bytes)
{
  char *mem = ti_region_get_mem(bytes);

#if 0 /* CM: No longer necessary. UNK_REGIONID is the default. */
  if (mem)
    set_region_range(mem, mem + bytes, UNK_REGIONID);
#endif

  return mem;
}
#endif /* USE_GC_NONE */

#ifdef MEMORY_DISTRIBUTED
void *ti_dist_malloc(size_t n)
{
  void *x = ti_malloc_atomic(n);

  if (x) {
    memset(x, 0, n);
  }

  return x;
}
#endif /* MEMORY_DISTRIBUTED */

/* Page to region map management */
/* ----------------------------- */

void set_region(struct page *p, int npages, Region r)
{
  PageId pnb = PAGENB(p);
  RegionId rid = Region2RegionId(r);

  while (npages-- > 0) 
    SetRegionId(pnb++, rid);
}

/* Mark the memory range from 'from' (inclusive) to 'to' (exclusive)
   as belonging to region with id 'rid' */
void set_region_range(void *from, void *to, RegionId rid)
{
  PageId first = PAGENB(from), last = PAGENB((PageId)to - 1);

  while (first <= last)
    SetRegionId(first++, rid);
}

#ifndef LARGE_ADDRESSES
/* This is a page table implemented as a simple table. */
RegionId __regionmap[MAXPAGE];
#else
/* This is a page table implemented as a table of indirect tables. */
RegionId *__regiontable[1 << MEMSLICE2];

void SetRegionId(PageId pagenb, RegionId rid)
{
  IntPtr offset1 = pagenb >> (MEMSLICE2 + MEMSLICE3);
  IntPtr offset2 = pagenb >> MEMSLICE3;
  IntPtr offset3 = pagenb & ((1 << MEMSLICE3) - 1);
  RegionId *rmap;

  /*assert(offset1 == 0);*/
  rmap = __regiontable[offset2];
  if (!rmap)
    {
      unsigned long size = sizeof(RegionId) * (1 << MEMSLICE3);
      int i;

#ifndef USE_GC_NONE
      if (size < GC_page_size) size = GC_page_size;
#endif
      /* assert(!(size & (GC_page_size - 1))); */

      rmap = (RegionId *) malloc(size);
      __regiontable[offset2] = rmap;
      /* assert((1 << MEMSLICE3) < sizeof(RegionId) &&
	 (((1 << MEMSLICE3) % sizeof(RegionId)) == 0)); */
      for (i = 0; i < (1 << MEMSLICE3); i++) {
	rmap[i] = UNK_REGIONID;
      }
    }
  rmap[offset3] = rid;
}
#endif

/* Multi-page allocation management */
/* -------------------------------- */

struct page *alloc_new(int n, struct page *next)
/* Assumes freepages_lock held */
{
  struct page *newp = (struct page *)ti_region_get_mem(n << RPAGELOG);

  if (!newp)
    abort();
  assert(!((long)newp & (RPAGESIZE - 1)));

  newp->next = next;
  newp->pagecount = n;
#ifndef NMEMDEBUG
  {
    PageId i, pnb = PAGENB(newp);

    for (i = pnb; i < pnb + n; i++)
      SetRegionId(i, FREEPAGE);
  }
#endif

  return newp;
}

struct page *alloc_split(struct page *split, int n, struct page *next)
/* Assumes freepages_lock held */
{
#ifndef NMEMDEBUG
  /* These pages had better be free */
  PageId i, pnb = PAGENB(split);

  assert(split->pagecount >= n);
  for (i = pnb; i < pnb + split->pagecount; i++)
    assert(Page2RegionId(i) == FREEPAGE);
#endif
  if (split->pagecount > n)
    {
      /* Keep first part of block */
      split->pagecount -= n;
      if (split->pagecount < K) /* move to solitary_pages */
	{
	  unlink_page(&unused_pages, split);
	  addfront(&solitary_pages, split);
	}
      /* Return latter part of block */
      split = (struct page *)((char *)split + (split->pagecount << RPAGELOG));
    }
  else
    {
      /* remove split from list */
      unlink_page(&unused_pages, split);
    }
  split->next = next;
  split->pagecount = n;

  return split;
}

struct page *alloc_pages(int n, struct page *next)
{
  struct page *best;
  jIntPointer bestn;
  struct page *scan;

  assert(n >= K);

#ifdef MEMORY_SHARED
  ti_mutex_lock(&freepages_lock);
#endif

  scan = unused_pages;
  /* Find first fit */
  for (;;)
    {
      if (!scan)
	{
	  struct page *newp = alloc_new(n, next);
#ifdef MEMORY_SHARED
	  ti_mutex_unlock(&freepages_lock);
#endif
	  return newp;
	}

      if (scan->pagecount >= n) break;
      scan = scan->next;
    }

  /* Now find best fit */
  best = scan;
  bestn = scan->pagecount;
  for (;;)
    {
      scan = scan->next;
      if (!scan)
	{
	  struct page *newp = alloc_split(best, n, next);

#ifdef MEMORY_SHARED
	  ti_mutex_unlock(&freepages_lock);
#endif
	  return newp;
	}

      if (scan->pagecount >=n && scan->pagecount < bestn)
	{
	  best = scan;
	  bestn = scan->pagecount;
	}
    }
}

void free_pages(Region r, struct page *p)
/* Assumes freepages_lock held */
{
#ifndef NMEMDEBUG
  PageId i, pnb = PAGENB(p);
  RegionId rid = Region2RegionId(r);

  for (i = pnb; i < pnb + p->pagecount; i++)
    {
      assert(Page2RegionId(i) == rid);
      SetRegionId(i, FREEPAGE);
    }
#endif

  addfront(&unused_pages, p);
  /* TBD: coalesce */
}


/* Single page management */
/* ---------------------- */
/* TBD: once in a while coalesce single_pages */

void add_single_pages(struct page *base, int n)
/* Effects: Adds the group of n pages at base to the single_pages list */
{
  while (n-- > 0)
    {
      ASSERT_FREE(base);
      base->next = single_pages;
      single_pages = base;
      base = (struct page *)((char *)base + RPAGESIZE);
    }
}

void scavenge_single_pages(int n)
{
  /* Add n pages to the single_pages list */
  struct page *scan, *best;
  jIntPointer bestn;

  /* First, take some solitary pages */
  while (solitary_pages)
    {
      /* Don't worry if we take a little too much here */
      struct page *adding = solitary_pages;

      solitary_pages = solitary_pages->next;
      if (solitary_pages) solitary_pages->previous = NULL;
      n -= adding->pagecount;
      add_single_pages(adding, adding->pagecount);
      if (n <= 0) return;
    }
  /* We still need pages. Take any group in unused_pages that is <= n.
     Remember smallest entry > n too. This is sortof equivalent to
     a best fit where we allow partial allocations to make up a whole */
  best = NULL;
  bestn = ULONG_MAX;
  scan = unused_pages;
  while (scan)
    {
      if (scan->pagecount <= n)
	{
	  struct page *adding = scan;

	  scan = scan->next;
	  n -= adding->pagecount;
	  unlink_page(&unused_pages, adding);
	  add_single_pages(adding, adding->pagecount);
	  if (n == 0) return;
	}
      else
	{
	  if (scan->pagecount < bestn)
	    {
	      bestn = scan->pagecount;
	      best = scan;
	    }
	  scan = scan->next;
	}
    }
  /* Still not enough. Split the best block if there is one, allocate
     new pages otherwise */
  if (!best)
    add_single_pages(alloc_new(n, NULL), n);
  else
    add_single_pages(alloc_split(best, n, NULL), n);
}

struct page *alloc_single_page(struct page *next)
{
  struct page *p;

#ifdef MEMORY_SHARED
  ti_mutex_lock(&freepages_lock);
#endif

  if (!single_pages)
    {
      scavenge_single_pages(PAGE_GROUP_SIZE);
    }
  ASSERT_FREE(single_pages);
  p = single_pages;
  single_pages = p->next;
  p->next = next;

#ifdef MEMORY_SHARED
  ti_mutex_unlock(&freepages_lock);
#endif

  return p;
}

void free_single_page(Region r, struct page *p)
/* Assumes freepages_lock held */
{
#ifndef NMEMDEBUG
  ASSERT_INUSE(p, r);
  SetRegionId(PAGENB(p), FREEPAGE);
#endif
  p->next = single_pages;
  single_pages = p;
}
