#ifndef __REGION
#define __REGION

/* This implementation assumes 32 bit pointers in the use of private
   region addresses as RegionId's and in calls to DEREF_GLOBAL to read
   remote pointers (Split-C backend) */

#include "config.h"
#include "backend.h"
#include "runtime-options.h"

#include <stdlib.h>
#include "linkage.h"
#include "region-struct.h"
#include "ti-memory.h"
#include "primitives.h"

#if SIZEOF_VOID_P == 8
#  define LARGE_ADDRESSES
#endif

typedef struct T6Region4lang2ti *Region;
typedef struct T6Region4lang2ti RegionStruct;

struct T12SharedRegion4lang2ti;
struct T13PrivateRegion4lang2ti;

/* CM: For RPAGESIZE, should we use the system page size?

   #include <unistd.h>
   long pagesize = sysconf(_SC_PAGESIZE);
*/

#define RPAGESIZE 8192
#define RPAGELOG 13

#ifdef WIDE_POINTERS

typedef GP_type(RegionStruct) GRegion;
typedef GP_type(jIntPointer) GP_jIntPointer;

#else
typedef Region GRegion;
#endif


#define FREEPAGE ((RegionId) -1L) /* Id of a free page */
#ifdef NMEMDEBUG
#define ASSERT_FREE(p) 0
#define ASSERT_INUSE(p, r) 0
#else
#define ASSERT_FREE(p) assert(PL2RegionId(p) == FREEPAGE)
#define ASSERT_INUSE(p, r) assert(PL2RegionId(p) == Region2RegionId((r)))
#endif

EXTERN_C_BEGIN

void region_init(void);
struct T12SharedRegion4lang2ti *newsregion(void);
struct T13PrivateRegion4lang2ti *newpregion(void);
Region newregion(void);
void initregion(Region r);
void *ralloc(Region r, size_t size);
const void *ralloc_atomic(Region r, size_t size);
int deleteregion(Region r);

EXTERN_C_END

#ifndef BACKEND_UNIPROC

/* On all backends but the uniprocessor one, region id's for private
   regions are just the address of the region, region id's of shared
   regions are odd numbers. The local representative of a shared region
   can be found (given the shared region id) by consulting the
   __sregionidmap table. Shared region id 1 is actually garbage
   collected memory. */

typedef jIntPointer RegionId;
#define UNK_REGIONID ((RegionId) 1)
#define SHARED_REGIONID(rid) ((rid) & 1)

extern Region STATIC_DEF(__sregionidmap[]);

static INLINE_MODIFIER Region RegionId2Region(RegionId rid)
{
  if (rid & 1) return STATIC_REF(__sregionidmap[rid >> 1]);
  else return (Region)rid;
}

#define Region2RegionId(r) ((r)->id)

#else  /* BACKEND_UNIPROC */

/* regions and region ids are the same thing */
typedef Region RegionId; 
extern struct T6Region4lang2ti zeroregion;
#define UNK_REGIONID &zeroregion
#define SHARED_REGIONID(rid) ((rid)->id & 1)

#define Region2RegionId(r) (r)
#define RegionId2Region(r) (r)

#endif /* BACKEND_UNIPROC */


typedef jUIntPointer PageId, IntPtr;

#ifndef LARGE_ADDRESSES

#define MAXMEMBITS 32
#define MAXPAGE (1 << (MAXMEMBITS - RPAGELOG))

extern RegionId __regionmap[];

static INLINE_MODIFIER void SetRegionId(PageId pagenb, RegionId rid)
{
  __regionmap[pagenb] = rid;
}

static INLINE_MODIFIER RegionId Page2RegionId(PageId pnb)
{
  return __regionmap[pnb];
}

#ifdef MEMORY_DISTRIBUTED

typedef GP_type(RegionId) GP_RegionId;

static INLINE_MODIFIER RegionId GPage2RegionId(int proc, PageId pnb)
{
  RegionId rv;
  GP_RegionId ga;

  TO_GLOBAL(ga, proc, __regionmap + pnb);
  DEREF_GLOBAL(rv, ga);

  return rv;
}

#endif

#else
/* We use a two-level "page table". The upper MEMSLICE1 bits are assumed
   to be 0. */

#define MAXMEMBITS 64
#define MEMSLICE1 15
#define MEMSLICE2 18
#define MEMSLICE3 18
#if (MEMSLICE1 + MEMSLICE2 + MEMSLICE3 + RPAGELOG != MAXMEMBITS)
#error "Bad address slices"
#endif

extern RegionId *__regiontable[];

void SetRegionId(PageId pagenb, RegionId rid);

static INLINE_MODIFIER RegionId Page2RegionId(PageId pnb)
{
  IntPtr offset1 = pnb >> (MEMSLICE2 + MEMSLICE3);
  IntPtr offset2 = pnb >> MEMSLICE3;
  IntPtr offset3 = pnb & ((1 << MEMSLICE3) - 1);

  /*assert(offset1 == 0);*/

  /* Check if indirect table is present */
  if (__regiontable[offset2]) {
    return __regiontable[offset2][offset3];
  } else {
    /* If the indirect table is not there, the value is assumed to be ... */
    return UNK_REGIONID;
  }
}

#ifdef MEMORY_DISTRIBUTED

/* This is slow. Things should be done differently (this probably involves
   moving the checking into the active message handler anyway, so will
   do this when I implement reference counting) */

typedef GP_type(RegionId) GP_RegionId;
typedef RegionId* RegionIdPtr;
typedef GP_type(RegionIdPtr) GP_RegionIdPtr;

static INLINE_MODIFIER RegionId GPage2RegionId(int proc, PageId pnb)
{
  IntPtr offset1 = pnb >> (MEMSLICE2 + MEMSLICE3);
  IntPtr offset2 = pnb >> MEMSLICE3;
  IntPtr offset3 = pnb & ((1 << MEMSLICE3) - 1);
  RegionId *level2;

  RegionId rv;
  GP_RegionId ga;
  GP_RegionIdPtr gap;

  /* assert(offset1 == 0); */

  TO_GLOBAL(gap, proc, __regiontable + offset2);
  DEREF_GLOBAL(level2, gap);
  /* Check if indirect table is present */
  if (level2) {
    TO_GLOBAL(ga, proc, level2 + offset3);
    DEREF_GLOBAL(rv, ga);

    return rv;
  } else {
    /* If the indirect table is not there, the value is assumed to be ... */
    return UNK_REGIONID;
  }
}

#endif

#endif

#define PAGENB(x) ((PageId)(x) >> RPAGELOG)
#define PAGEADDR(n) ((void *)((n) << RPAGELOG))

void set_region(struct page *p, int npages, Region r);
/* Mark the memory range from 'from' (inclusive) to 'to' (exclusive)
   as belonging to region with id 'rid' */
void set_region_range(void *from, void *to, RegionId rid);

static INLINE_MODIFIER RegionId PL2RegionId(const void *p)
{
  return Page2RegionId(PAGENB(p));
}

static INLINE_MODIFIER Region PL2Region(const void *p)
{
  return RegionId2Region(PL2RegionId(p));
}


#ifdef MEMORY_DISTRIBUTED

# define MyRegion(r) (r)

static INLINE_MODIFIER int GRegion2RegionId(GRegion r)
{
  jIntPointer id;
  GP_jIntPointer idAddr;

  FIELD_ADDR_GLOBAL(idAddr, r, id);
  DEREF_GLOBAL(id, idAddr);
  return id;
}

# define PG2RegionId(p)  split2RegionId(TO_PROC(p), TO_LOCAL(p))

static INLINE_MODIFIER RegionId split2RegionId(int proc, const void *addr)
{
  if (proc == MYPROC || !addr) return PL2RegionId(addr);
  else return GPage2RegionId(proc, PAGENB(addr));
}

# define PG2Region(p)  split2Region(TO_PROC(p), TO_LOCAL(p))

static INLINE_MODIFIER Region
split2Region(int proc, const void *addr)
{
  return RegionId2Region(split2RegionId(proc, addr));
}

# define MyGRegion(x) MyGRegionHandler(TO_PROC(x), (Region) (TO_LOCAL(x)))

static INLINE_MODIFIER Region MyGRegionHandler(int proc, Region r)
{
  if (proc == MYPROC)
    return r;

  /* Must be a shared region (otherwise we wouldn't be seeing it) */
  {
    GRegion gp_r;
    TO_GLOBAL(gp_r, proc, r);
    return __sregionidmap[GRegion2RegionId(gp_r) >> 1];
  }
}

#else /* !MEMORY_DISTRIBUTED */
/* uniproc, smp-narrow, smp-wide, and pthread backends */

# ifdef MEMORY_SHARED

static INLINE_MODIFIER Region MyRegion(Region r)
{
  if (!r) return NULL;
  /* Returns our representative for region r */
  return RegionId2Region(r->id);
}

# else  /* !MEMORY_SHARED */

/* uniproc backend */
#  define MyRegion(r) (r)

# endif /* !MEMORY_SHARED */

# ifdef WIDE_POINTERS
#  define GRegion2RegionId(x) Region2RegionId((Region)TO_LOCAL(x))
#  define PG2RegionId(x) PL2RegionId((Region)TO_LOCAL(x))
#  define PG2Region(x) PL2Region((Region)TO_LOCAL(x))
#  define MyGRegion(x) MyRegion((Region)TO_LOCAL(x))
# else /* !WIDE_POINTERS */
#  define GRegion2RegionId Region2RegionId
#  define PG2RegionId PL2RegionId
#  define PG2Region PL2Region
#  define MyGRegion(x) MyRegion((GRegion)(x))
# endif/* !WIDE_POINTERS */

#endif /* !MEMORY_DISTRIBUTED */

void sharingViolation(void);

#endif /* !__REGION */
